//	KaleidoPaintApp.swift
//
//	© 2025 by Jeff Weeks
//	See TermsOfUse.txt

import SwiftUI

@main
struct KaleidoPaintApp: App {

	@Environment(\.scenePhase) var scenePhase

	@State var quickStartHelpShouldBeShown: Bool = true
	@State var editorHelpShouldBeShown: Bool = true
	
	init() {

		UserDefaults.standard.register(
			defaults: [

				//	At what initial resolution should we create new drawings?
				//
				//		Low		2⁹  × 2⁹  =  512 ×  512 pixels
				//		Medium	2¹⁰ × 2¹⁰ = 1024 × 1024 pixels
				//		High	2¹¹ × 2¹¹ = 2048 × 2048 pixels
				//
				"default log₂(unit cell size)" : 10,
				
				//	With what initial symmetry group should we create new drawings?
				"default symmetry group" : "632"
			]
		)

		if !UserDefaults.standard.bool(forKey: "legacy files have been converted") {
			ConvertLegacyFiles()
		}

#if targetEnvironment(simulator)
		//	When creating screenshots in the simulator,
		//	we'll need to know where to put source files
		//	for the desired drawings.
		PrintDocumentDirectoryPathToConsole()
#endif
	}
	
	var body: some Scene {

		DocumentGroup(newDocument: { KaleidoPaintDocument() }) { file in

			let theRenderer = ((gMeasureDrawingArea || gMakeScreenshots) ?
								nil :
								KaleidoPaintRenderer(log2UnitCellResolution:
									file.document.itsLog2UnitCellResolution))

			KaleidoPaintContentView(
				document: file.document,
				fileURL: file.fileURL,
				renderer: theRenderer,
				quickStartHelpShouldBeShown: $quickStartHelpShouldBeShown,
				editorHelpShouldBeShown: $editorHelpShouldBeShown)
		}
		.onChange(of: scenePhase) { newPhase in
		
			//	Each time KaleidoPaint comes out of background...
			if newPhase == .background {
			
				//	...we'll want to show the .helpQuickStart panel
				//	(if and when the user opens a new drawing)
				//	and the .helpEditor panel (if and when
				//	the user enters Edit mode).
				
				quickStartHelpShouldBeShown = true
				editorHelpShouldBeShown = true
			}
		}
	}
}

func ConvertLegacyFiles() {

	do {
		let theFileManager = FileManager.default
		
		//	Get a reference to the app's Documents directory.
		let theDocumentsDirectory = try theFileManager.url(
				for: .documentDirectory,
				in: .userDomainMask,
				appropriateFor: nil,
				create: true)
	
		//	What does the Documents directory contain?
		let theDirectoryContents = try theFileManager.contentsOfDirectory(
			at: theDocumentsDirectory,
			includingPropertiesForKeys: [.isRegularFileKey, .nameKey],
			options: [.skipsHiddenFiles])

		//	Consider each file...
		for theFileURL in theDirectoryContents {
			
			do {
				
				let theResourceValues = try theFileURL.resourceValues(forKeys: [.nameKey, .isRegularFileKey])
				if !(theResourceValues.isRegularFile ?? false) {
					continue
				}
				guard let theFileName = theResourceValues.name else {
					continue
				}

				//	If this file is a drawing file,
				//	change its filename extension from ".txt" to ".kaleidopaint" .
				//
				//		Note:  In the unlikely event that the user has placed
				//		some other .txt file (not a KaleidoPaint drawing) into the directory,
				//		KaleidoPaintThumbnailProvider's provideThumbnail() method
				//		will notice that decodeFileFormat() fails, and will call
				//
				//			handler(nil, KaleidoPaintThumbnailError.invalidDrawingData)
				//
				//		to report the problem to the document picker.
				//		The document picker thus knows to keep that file disabled,
				//		and not let the user open it in KaleidoPaint.
				//
				if theFileName.hasSuffix(".txt") {
					let theNewFileURL = theFileURL.deletingPathExtension().appendingPathExtension("kaleidopaint")
					try theFileManager.moveItem(at: theFileURL, to: theNewFileURL)
				}
				
				//	If this file is the drawings manifest or a thumbnail image,
				//	delete it.
				if (theFileName == "KaleidoPaintPaintingsManifest.xml"
				 || theFileName.hasSuffix(".png")
				) {
					try theFileManager.removeItem(at: theFileURL)
				}

			} catch {
				//	In the unlikely event that we failed to delete/rename a file,
				//	keep going in hopes of deleting/renaming the other files.
			}
		}
		
		//	The legacy files have now all been renamed/deleted.
		//	For this user on this device, there's no need
		//	to call ConvertLegacyFiles() ever again.
		UserDefaults.standard.set(true, forKey:"legacy files have been converted")

	} catch {
		
		//	In the unlikely case that something went wrong,
		//	let the user push onward without converting the files.
		//	We'll try the conversion again the next time the user launches the app.
	}
}
