//	KaleidoPaintContentView.swift
//
//	© 2025 by Jeff Weeks
//	See TermsOfUse.txt

import SwiftUI
import simd
import CoreMedia


enum PanelType {
	case none
	case brushColor
	case fillColor
	case brushThickness
	case smoothingTolerance
	case backgroundColor
	case symmetry
	case export
	case helpQuickStart
	case helpEditor
	case helpTutorials
	case helpImageGallery
	case helpContact
	case translators
}

enum SelectionMode: CaseIterable, Identifiable {

	case new		//	select a new color or thickness
	case previous	//	choose from among the colors or thicknesses already in use in the drawing

	var id: Self {self}

	func symbolSystemName() -> String {
		
		switch self {
		
		case .new:
			return "rays"
			
		case .previous:
			return "clock.arrow.circlepath"
		}
	}
}


let toolbarPadding: CGFloat = 8.0
let selectedButtonCornerRadius: CGFloat = geometryGamesCornerRadius - toolbarPadding
let selectedButtonBackgroundColor: Color = Color(white: 0.875)

//	At log2UnitCellResolution = 9 (low resolution) a sample drawing
//	that Amber sent leaks (with 4-way flood fill) at -2.906769,
//	but stops leaking at -2.895961.  Given that my own fairly extensive
//	leak-test sample drawing stopped leaking at -2.98, it's clear
//	that the thickness at which leakage occurs can vary from one drawing
//	to another.  So I'm raising the minimum to -2.8 to allow a little slack.
//
//	With my own leak-test sample drawing, here's the comparison
//	between 4-way and 8-way flood fills:
//
//		With 8-way flood fill
//			-2.51	already leaks (might leak even for thicker lines)
//
//		With 4-way flood fill
//			-2.98	no leakage detected in limited testing
//			-2.99	just a little leakage
//			-3.01	lots of leakage
//
let minLog2ThicknessFactor = -2.8	//	allows some slack relative to -2.895961 value noted above;
									//	somewhat more than 1/8 of the default thickness
let maxLog2ThicknessFactor = +2.0	//	4 times the default thickness
let thicknessSliderLength = 256.0

//	gMakeScreenshots is defined in KaleidoPaintRenderer.swift
//	rather than here, because the KaleidoPaintThumbnailProvider
//	also needs to see it.

//	Enable gMeasureDrawingArea to measure the area available
//	to the KaleidoPaintViewRep, which is the whole display
//	except for the navigation bar at the top.  As of March 2023,
//	the sizes required for the App Store are
//
//		5.5" device - iPhone 8 Plus
//			display size:  1242 × 2208
//			drawing size:  1242 × 2076 (with status bar suppressed)
//
//		6.7" device - iPhone 14 Pro Max
//			display size:  1290 × 2796
//			drawing size:  1290 × 2503 (with status bar suppressed)
//
//		12.9" device - iPad Pro (2nd gen)
//			display size:  2048 × 2732
//			drawing size:  2048 × 2632 (with status bar suppressed)
//
//		12.9" device - iPad Pro (6th gen)
//			display size:  2048 × 2732
//			drawing size:  2048 × 2632 (with status bar suppressed)
//
//#warning("disable gMeasureDrawingArea")
let gMeasureDrawingArea = false

struct KaleidoPaintContentView: View {

	@Environment(\.horizontalSizeClass) var horizontalSizeClass
	@Environment(\.undoManager) var defaultUndoManager

	@ObservedObject var document: KaleidoPaintDocument
	let fileURL: URL?
	
	//	A renderer is a potentially expensive resource,
	//	so let's let the KaleidoPaintView and the GeometryGamesExportView
	//	share a single renderer, instead of forcing each to create its own.
	//
	//	I tried various ways of setting this up a @State or @StateObject,
	//	but I couldn't persuade SwiftUI to rebuild the affected Views
	//	when I replaced the KaleidoPaintRenderer.  So I resorted
	//	to the embarrassing hack of incrementing itsID.
	//
	//		Note:  Even though this is a @State property and
	//		not a @StateObject property, different KaleidoPaintContentViews
	//		(in different scenes) each get their own KaleidoPaintRenderer.
	//
	@State var renderer: KaleidoPaintRenderer?
	@State var itsID = 0

	@State var touchMode: TouchMode = .draw

	@State var visiblePanel: PanelType = .none

	//	Let the SelectionMode and "Change all of…" settings
	//	persist even when the user closes and reopens
	//	a ColorPanel or ThicknessPanel.
	@State var brushColorSelectionMode = SelectionMode.new
	@State var brushThicknessSelectionMode = SelectionMode.new
	@State var fillColorSelectionMode = SelectionMode.new
	@State var backgroundColorSelectionMode = SelectionMode.new
	@State var editorColorSelectionMode = SelectionMode.new
	@State var editorChangeAllOfSameColor = false
	@State var editorThicknessSelectionMode = SelectionMode.new
	@State var editorChangeAllOfSameThickness = false
	@State var editorDeleteAllOfSameColor = false

	//	Auto-show the helpQuickStart and helpEditor panels once per session.
	@Binding var quickStartHelpShouldBeShown: Bool
	@Binding var editorHelpShouldBeShown: Bool
	
	//	Legacy drawings may contain Curves that are thinner
	//	than KaleidoPaint 3.0's minLog2ThicknessFactor allows.
	//	Flood-fill paint may leak through such Curves,
	//	so when opening a drawing we'll automatically thicken them.
	//	This operation could have side effects -- for example
	//	if the user had place a fill-point very close to one
	//	of those too-thin lines, the flood-fill point could
	//	end up flowing into the Curve itself rather than
	//	into the nearby region.  Ideally users will correct
	//	any such problems and keep the thickened lines,
	//	but we should at least give them the option to undo
	//	the thickening, and display an Alert explaining
	//	the thickening and saying how to undo it.
	//	And in any case, we want to respect the principle
	//	of not modifying a user's drawings without his/her consent.
	@State var shouldTestForTooThinLinesWhenUnlocked = false	//	overridden in onAppear
	@State var autoThickeningMessageShouldBeShown = false		//	overridden in onAppear
	


	var body: some View {

		//	Our KaleidoPaintDocument is a ReferenceFileDocument
		//	which, when we close the document, will save its contents
		//	iff there are un-doable changes sitting on the UndoManager
		//	that it passes to us as an @Environment variable.
		//	So it's essential that our document use that default UndoManager
		//	and not one that we allocate ourselves.
		if let theDefaultUndoManager = defaultUndoManager {
			document.itsUndoManager = theDefaultUndoManager
		} else {
			assertionFailure("The defaultUndoManager is missing")
		}
	
		//	When I someday require iOS 17.1 or higher,
		//	I may remove the following workaround
		//	for the iOS 17.0.x SwiftUI document-based app nav bar bug.
		let theIOSVersion = ProcessInfo.processInfo.operatingSystemVersion
		let theSwiftUIBugIsPresent = (
			theIOSVersion.majorVersion == 17
		 && theIOSVersion.minorVersion == 0 )

		//	This top-level ZStack lets the controls overlay the main view.
		return ZStack(alignment: .center) {

			//	Main animation view
			if let theRenderer = renderer {

				KaleidoPaintViewRep(
					document: document,
					renderer: theRenderer,
					visiblePanel: $visiblePanel,
					touchMode: $touchMode)
				.edgesIgnoringSafeArea(.bottom)	//	set to .all to underlap nav bar as well
				.popover(
					isPresented: $document.itsEditorIsShown,
					attachmentAnchor: document.itsEditorAnchor
					) {
						EditorPanel(
							document: document,
							colorSelectionMode: $editorColorSelectionMode,
							changeAllOfSameColor: $editorChangeAllOfSameColor,
							thicknessSelectionMode: $editorThicknessSelectionMode,
							changeAllOfSameThickness: $editorChangeAllOfSameThickness,
							deleteAllOfSameColor: $editorDeleteAllOfSameColor)
					}
				.onChange(of: document.itsEditorIsShown) { editorIsShown in

					if !editorIsShown {

						//	The user dismissed the Editor popover,
						//	so unselect the selection.

						document.itsControlHandleHit = nil

						document.itsBaseDrawingTextureNeedsFullRedraw = true
						document.itsModelChangeCount += 1
					}
				}
			}

			//	If we ever need to make screenshots for a new size device,
			//	and the Simulator still won't KaleidoPaint's graphics code,
			//	then...
			//
			//	1.  We'll need to enable gMeasureDrawingArea just long enough
			//	to learn how big the KaleidoPaintView will be on the new device.
			//
			if gMeasureDrawingArea {

				ZStack() {
				
					Rectangle()
					.foregroundColor(.red)

					GeometryReader {geo in
						VStack() {
							Text("\(    geo.size.width) × \(    geo.size.height)")
							Text("\(2 * geo.size.width) × \(2 * geo.size.height)")
							Text("\(3 * geo.size.width) × \(3 * geo.size.height)")
						}
						.frame(maxWidth: .infinity, maxHeight: .infinity, alignment: .center)
					}
				}
				.edgesIgnoringSafeArea(.bottom)
			}
			//
			//	2.  We'll need to render the desired drawings
			//	on an actual device and export them at the desired sizes.
			//	See "makeKaleidoPaintScreenshotImages" in GeometryGamesExportView.swift.
			//
			//	3.  We'll need to display those pre-rendered images here.
			//
			if gMakeScreenshots {
			
				if let theDrawingURL = fileURL {
				
					GeometryReader() { reader in
					
						screenshotImage(
							drawingURL: theDrawingURL,
							height: reader.size.height)
						.frame(maxWidth: .infinity, alignment: .center)
					}
					.edgesIgnoringSafeArea(.bottom)
				}
			}
			
			//	Controls
			VStack(alignment: .center) {
			
				ButtonsAtTop(
					document: document,
					fileURL: fileURL,
					shouldTestForTooThinLinesWhenUnlocked: $shouldTestForTooThinLinesWhenUnlocked,
					autoThickeningMessageShouldBeShown: $autoThickeningMessageShouldBeShown
				)
				.onChange(of: document.itsLog2UnitCellResolution) { newLog2UnitCellSize in
					
					//	Clear all cached RLE runs.
					document.clearCachedRLERuns()
					
					//	Replace the renderer with a new one using the new unit cell size.
					renderer = KaleidoPaintRenderer(log2UnitCellResolution: newLog2UnitCellSize)
					itsID += 1	//	ugly hack to force SwiftUI to rebuild the affected Views
					
					//	Request a redraw.

					document.itsBaseDrawingTextureNeedsFullRedraw = true
					document.itsCurveInProgressTextureNeedsRedraw = true
					document.itsSingularitiesTextureNeedsRedraw = true
					document.itsCompositeTextureNeedsRefresh = true

					document.itsModelChangeCount += 1
				}
				.onChange(of: document.itsContentIsLocked) { contentGotLocked in

					if contentGotLocked {
					
						if touchMode != .draw {
							touchMode = .draw
							document.itsBaseDrawingTextureNeedsFullRedraw = true
							document.itsModelChangeCount += 1
						}

						visiblePanel = .none

						//	When the user locks the drawing, I would have
						//	preferred to leave the Undo and Redo stacks untouched,
						//	and simply disable the Undo/Redo buttons.  That way,
						//	if the user later unlocked the drawing (before
						//	closing it), all those changes would still be available
						//	on the Undo/Redo stacks.  But even though I can
						//	disable KaleidoPaint's own Undo/Redo buttons,
						//	there doesn't seem to be any way to disable SwiftUI's
						//	built-in 3-finger swipe left / swipe right gestures,
						//	which also trigger undo/redo.  The only way
						//	I could find to stop the user from modifying
						//	a locked drawing via a 3-finger swipe was
						//	to clear the Undo/Redo stacks completely.
						//	This isn't a perfect solution, but it's
						//	the best I could find.
						//
						document.itsUndoManager?.removeAllActions()
						document.itsUndoRedoCount += 1
					}
				}

				Spacer()
				
				if horizontalSizeClass == .regular
				|| touchMode == .edit {
				
					if let theRenderer = renderer {
					
						PossibleControlPanel(
							document: document,
							visiblePanel: $visiblePanel,
							touchMode: $touchMode,
							brushColorSelectionMode: $brushColorSelectionMode,
							brushThicknessSelectionMode: $brushThicknessSelectionMode,
							fillColorSelectionMode: $fillColorSelectionMode,
							backgroundColorSelectionMode: $backgroundColorSelectionMode,
							quickStartHelpShouldBeShown: $quickStartHelpShouldBeShown,
							fileURL: fileURL,
							renderer: theRenderer)
					}
					
					//	This ZStack lets the TouchModePicker be perfectly centered.
					ZStack(alignment: .bottom) {

						if !document.itsContentIsLocked {
						
							TouchModePicker(
								document: document,
								visiblePanel: $visiblePanel,
								editorHelpShouldBeShown: $editorHelpShouldBeShown,
								touchMode: $touchMode)
						}
		
						HStack(alignment: .bottom) {

							if !document.itsContentIsLocked
							&& touchMode == .draw {
							
								ButtonsAtBottomLeading(
									document: document,
									visiblePanel: $visiblePanel)
							}
							
							Spacer()
						
							ButtonsAtBottomTrailing(
								document: document,
								visiblePanel: $visiblePanel,
								touchMode: $touchMode)
						}
					}
					
				} else {	//	horizontalSizeClass == .compact (or .none or .some ???) and
							//	touchMode == .draw

					//	This ZStack lets the TouchModePicker be perfectly centered.
					ZStack(alignment: .bottom) {

						//	This is the real TouchModePicker.
						if !document.itsContentIsLocked {

							TouchModePicker(
								document: document,
								visiblePanel: $visiblePanel,
								editorHelpShouldBeShown: $editorHelpShouldBeShown,
								touchMode: $touchMode)
						}

						HStack(alignment: .bottom, spacing: 0.0) {

							if !document.itsContentIsLocked {
							
								ButtonsAtBottomLeading(
									document: document,
									visiblePanel: $visiblePanel)
							}

							VStack() {

								PossibleControlPanel(
									document: document,
									visiblePanel: $visiblePanel,
									touchMode: $touchMode,
									brushColorSelectionMode: $brushColorSelectionMode,
									brushThicknessSelectionMode: $brushThicknessSelectionMode,
									fillColorSelectionMode: $fillColorSelectionMode,
									backgroundColorSelectionMode: $backgroundColorSelectionMode,
									quickStartHelpShouldBeShown:
										$quickStartHelpShouldBeShown,
									fileURL: fileURL,
									renderer: renderer)
								.padding(EdgeInsets(
									top: 0.0,
									leading: geometryGamesPanelMargin,
									bottom: 0.0,
									trailing: geometryGamesPanelMargin))

								//	This is a dummy TouchModePicker.
								//	It won't be perfectly centered horizontally (because
								//	ButtonsAtBottomLeading and ButtonsAtBottomTrailing()
								//	don't have exactly the same width), but vertically
								//	it will force the PossibleControlPanel() to the desired height.
								//
								if !document.itsContentIsLocked {
								
									TouchModePicker(
										document: document,
										visiblePanel: $visiblePanel,
										editorHelpShouldBeShown: $editorHelpShouldBeShown,
										touchMode: $touchMode)
									.hidden()
								}
							}

							ButtonsAtBottomTrailing(
								document: document,
								visiblePanel: $visiblePanel,
								touchMode: $touchMode)
						}
					}
				}
			}
			.padding(geometryGamesControlInset)
			.onAppear() {
				
				if gMakeScreenshots {
				
					visiblePanel = selectPanelForScreenshot(fileURL: fileURL)
					
					brushColorSelectionMode = .previous	//	used with brush color panel in "Cubic"
					
				} else
				{
					if quickStartHelpShouldBeShown
					&& touchMode == .draw
					&& visiblePanel == .none {
					
						visiblePanel = .helpQuickStart
					}
				}
				
				//	Put a "null change" onto the Undo stack,
				//	so that the drawing will get re-saved when the user
				//	closes it, even if the user has made only non-undoable
				//	changes like rotating, zooming, or locking the drawing.
				//	It's not foolproof:  if the user undoes this null change
				//	(and makes no other undoable changes) then the drawing
				//	won't get resaved and any non-undoable changes will be lost.
				//
				document.changeNothing()

				if document.itsContentIsLocked {
				
					//	Wait until later, when the user unlocks the drawing,
					//	to test for too-thin lines.
					
					shouldTestForTooThinLinesWhenUnlocked = true
					autoThickeningMessageShouldBeShown = false
					
				} else {
				
					//	The drawing is unlocked,
					//	so test for too-thin lines now.
				
					shouldTestForTooThinLinesWhenUnlocked = false
					autoThickeningMessageShouldBeShown
						= autoThickenBrushstrokesInLegacyDrawing(document: document)
				}
			}
			.alert("AutoThickenMessage", isPresented: $autoThickeningMessageShouldBeShown) {
				Button("OK", role: .cancel) { }
			}
			
			if theSwiftUIBugIsPresent {
			
				VStack(spacing: 8.0) {
					
					Text("iOS 17.1 required")
					.font(.title)
					.foregroundColor(.red)

					Text("KaleidoPaint runs correctly on iOS 15, 16 and 17.1 or later. But it won't run correctly on iOS 17.0.x because a bug in iOS 17.0.x fails to show the Back button and the drawing title in SwiftUI apps. Please update your device to iOS 17.1 or later to solve this problem.")
					.font(.title2)
					.foregroundColor(.black)
				}
				.padding(24.0)
				.background(.white)
				.cornerRadius(16.0)
				.overlay(
					RoundedRectangle(cornerRadius: 16.0)
					.stroke(.black, lineWidth: 4.0)
				)
				.padding(32.0)
			}
		}
		.id(itsID)
		//	When most users are running iOS 17 or later
		//	we may eliminate DrawingTitleStyle entirely
		//	(assuming no more surprises in iOS 18 or later).
		.modifier(DrawingTitleStyle())
	}
	
	struct DrawingTitleStyle: ViewModifier {
	
		func body(content: Content) -> some View {
	
			if #available(iOS 17, *) {
			
				//	On iOS 17, the title is displayed .inline by default.
				content
			
			} else if #available(iOS 16, *) {

				//	On iOS 16, the title is displayed .large by default.
				//	Override that default to display an .inline title.
				content
				.navigationBarTitleDisplayMode(.inline)

			} else {

				//	On iOS 15, navigationBarTitleDisplayMode(.inline)
				//	fails to show the title at all (!).  Fortunately
				//	if we omit navigationBarTitleDisplayMode(),
				//	then the title shows as desired.
				content
			}
		}
	}
}

@MainActor func autoThickenBrushstrokesInLegacyDrawing(
	document: KaleidoPaintDocument
) -> Bool {	//	were changes made?

	let theSafeThickness = exp2(minLog2ThicknessFactor) * gDefaultThickness	//	currently 0.002243…
	
	//	Thicken the paintbrush if it's too thin.
	if document.itsBrushThickness < theSafeThickness {
		document.itsBrushThickness = theSafeThickness
	}
	
	//	Thicken any too-thin Brushstrokes.
	
	var theNumThickenedLines = 0
	
	let theOldThicknesses = document.itsElements.map() { element -> Double? in
		if let theBrushstroke = element as? Brushstroke {
			return theBrushstroke.itsThickness
		} else {
			return nil
		}
	}

	let theNewThicknesses = theOldThicknesses.map() { thickness -> Double? in
		if let theThickness = thickness {
			if theThickness < theSafeThickness {
				theNumThickenedLines += 1
				return theSafeThickness
			} else {
				return thickness	//	= theThickness here
			}
		} else {
			return thickness		//	= nil here
		}
	}
		
	if theNumThickenedLines > 0 {

		document.changeLegacyDrawingBrushstrokeThicknesses(
			from: theOldThicknesses,
			to: theNewThicknesses)

		return true
		
	} else {
	
		return false	//	No changes were made, no need to display a message.
	}
}


// MARK: -
// MARK: Drawing view

struct KaleidoPaintViewRep: UIViewRepresentable {

	//	Someday SwiftUI might offer direct access
	//	to a CAMetalLayer or equivalent. Until then, we define
	//	a KaleidoPaintView to be a subclass of a UIView,
	//	and wrap an instance of one in this KaleidoPaintViewRep,
	//	which implements the UIViewRepresentable protocol
	//	that SwiftUI requires for including a UIView
	//	in the view heirarchy.
	
	//	How iOS uses this struct at runtime is a bit surprising.
	//	When the user first launches the app, we get calls to
	//
	//		init()
	//		makeUIView()
	//		updateUIView()
	//
	//	as expected.  But thereafter, whenever
	//	the KaleidoPaintContentView's body() gets called,
	//	here we get only a call to init(),
	//	with no call to makeUIView() nor updateUIView().
	//	That is, iOS knows that it can keep using
	//	the same KaleidoPaintView that it's already got,
	//	so it doesn't ask for a new one, even though
	//	it's created a new KaleidoPaintViewRep "just in case".
	//	In effect, the KaleidoPaintViewRep is a factory
	//	that can be used to create a new KaleidoPaintView,
	//	but in most situations it's not needed.

	var document: KaleidoPaintDocument	//	no need for @ObservedObject, because the animation
										//		responds to changes in itsModelChangeCount
	var renderer: KaleidoPaintRenderer
	@Binding var visiblePanel: PanelType
	@Binding var touchMode: TouchMode
	
	func makeUIView(
		context: UIViewRepresentableContext<KaleidoPaintViewRep>
	) -> KaleidoPaintView {

		let theDrawingView = KaleidoPaintView(
			frame: CGRect.zero,
			modelData: document,
			renderer: renderer,
			visiblePanel: $visiblePanel,
			touchMode: $touchMode)

		return theDrawingView
	}
	
	func updateUIView(
		_ uiView: KaleidoPaintView,
		context: UIViewRepresentableContext<KaleidoPaintViewRep>
	) {
	}
}


// MARK: -
// MARK: Screenshots on simulator

func screenshotImage(
	drawingURL: URL,
	height: Double
) -> Image {

	let theExtensionlessURL = drawingURL.deletingPathExtension()
	let theDrawingName = theExtensionlessURL.lastPathComponent
	let theFolderURL = theExtensionlessURL.deletingLastPathComponent()
	
	let theSizeSuffix: String
	switch Int(floor(height)) {
	
		case  692:								//	points, not pixels
			theSizeSuffix = " 1242 × 2076@3x"	//	pixels, not points

		//	On the iPhone 14 Pro Max simulator we get called
		//	first with the surprising height of 888.0 points,
		//	then immediately afterwards we get called again
		//	with the expected height of 834.333… points.
		//	I'm not sure what's going on.  But this is just
		//	screenshot code, not release code, so I guess
		//	it's OK.
		case  888, 834:	//	= 834.333…			//	points, not pixels
			theSizeSuffix = " 1290 × 2503@3x"	//	pixels, not points
	
		case 1316:								//	points, not pixels
			theSizeSuffix = " 2048 × 2632@2x"	//	pixels, not points
	
		default:
			preconditionFailure("unexpected view height")
	}

	let theImageURL = theFolderURL.appendingPathComponent(
//							  theDrawingName	//	used as folder name
//							+ "/"
//							+
							  theDrawingName	//	used as part of file name
							+ theSizeSuffix
							+ ".png",
						isDirectory: false)
	
	//	Hmmm... let's use a UIImage here.
	//	I'm not sure how to do this with pure SwiftUI.
	guard let theUIImage = UIImage(contentsOfFile: theImageURL.relativePath) else {
		preconditionFailure("couldn't read file")
	}
	let theImage = Image(uiImage: theUIImage)

	return theImage
}

func selectPanelForScreenshot(
	fileURL: URL?
) -> PanelType {

	if let theFileURL = fileURL {
	
		let theDrawingName = theFileURL.deletingPathExtension().lastPathComponent
		switch theDrawingName {
		
			case "Paisley":
				return .helpQuickStart
				
			case "Cubic":
				return .brushColor
				
			case "Knitted":
				return .symmetry
			
			case "Lilylike":
				return .helpEditor
			
			case "Eagles":
				return .smoothingTolerance
		
			default:
				return .none
		}
		
	} else {	//	should never occur
	
		return .none
	}
}



// MARK: -
// MARK: Toolbars

struct BorderIfNeeded: ViewModifier {

	@Environment(\.colorScheme) var colorScheme

	let backgroundIntensity: Double
	let cornerRadius: Double
	
	init(
		over backgroundColor: LinearDisplayP3Color,
		cornerRadius: Double
	) {
		self.backgroundIntensity
			= (backgroundColor.r
			 + backgroundColor.g
			 + backgroundColor.b)
			/ 3.0
			
		self.cornerRadius = cornerRadius
	}
	
	func body(content: Content) -> some View {
	
		let theBorderIsNeeded = (
			colorScheme == .light ?
			(backgroundIntensity > 0.9375) :
			(backgroundIntensity < 0.0625)
		)
		
		return Group() {
		
			if theBorderIsNeeded {
			
				content
				.overlay(
					RoundedRectangle(cornerRadius: cornerRadius)
					.stroke(.gray, lineWidth: 1.0))

			} else {
			
				content
			}
		}
	}
}

struct ButtonsAtTop: View {

	@ObservedObject var document: KaleidoPaintDocument
	let fileURL: URL?
	@Binding var shouldTestForTooThinLinesWhenUnlocked: Bool
	@Binding var autoThickeningMessageShouldBeShown: Bool

	var body: some View {

		//	This ZStack lets the Undo and Redo buttons be centered.
		ZStack(alignment: .top) {

			ResolutionButton(document: document)
			.frame(maxWidth: .infinity, alignment: .leading)

			if !document.itsContentIsLocked {
			
				HStack() {
					UndoButton(document: document)
					RedoButton(document: document)
				}
				.padding(toolbarPadding)
				.background(Color(UIColor.systemBackground))
				.cornerRadius(geometryGamesCornerRadius)
				.modifier(BorderIfNeeded(
					over: document.itsBackgroundColor,
					cornerRadius: geometryGamesCornerRadius))
			}

			LockButton(
				document: document,
				fileURL: fileURL,
				shouldTestForTooThinLinesWhenUnlocked: $shouldTestForTooThinLinesWhenUnlocked,
				autoThickeningMessageShouldBeShown: $autoThickeningMessageShouldBeShown)
			.frame(maxWidth: .infinity, alignment: .trailing)
		}
	}
}

struct ButtonsAtBottomLeading: View {

	@ObservedObject var document: KaleidoPaintDocument
	@Binding var visiblePanel: PanelType

	var body: some View {

		//	The buttons already have their own tappable padding,
		//	so the AdaptiveStack doesn't need to provide any extra spacing.
		AdaptiveStack(spacing: 0) {

			BrushColorButton(
				visiblePanel: $visiblePanel,
				brushColor: $document.itsBrushColor)

			FillColorButton(
				visiblePanel: $visiblePanel,
				fillColor: $document.itsFillColor)

			BrushThicknessButton(
				visiblePanel: $visiblePanel)

			SmoothingToleranceButton(
				visiblePanel: $visiblePanel)
		}
		.padding(toolbarPadding)
		.background(Color(UIColor.systemBackground))
		.cornerRadius(geometryGamesCornerRadius)
		.modifier(BorderIfNeeded(
			over: document.itsBackgroundColor,
			cornerRadius: geometryGamesCornerRadius))
	}
}

struct ButtonsAtBottomTrailing: View {

	@ObservedObject var document: KaleidoPaintDocument
	@Binding var visiblePanel: PanelType
	@Binding var touchMode: TouchMode

	var body: some View {

		//	The buttons already have their own tappable padding,
		//	so the AdaptiveStack doesn't need to provide any extra spacing.
		AdaptiveStack(spacing: 0) {

			if !document.itsContentIsLocked
			&& touchMode == .draw {
			
				BackgroundColorButton(
					visiblePanel: $visiblePanel,
					backgroundColor: $document.itsBackgroundColor)

				SymmetryButton(
					visiblePanel: $visiblePanel,
					wallpaperGroup: $document.itsWallpaperGroup)
			}

			ExportButton(
				visiblePanel: $visiblePanel)

			HelpButton(
				visiblePanel: $visiblePanel)
		}
		.padding(toolbarPadding)
		.background(Color(UIColor.systemBackground))
		.cornerRadius(geometryGamesCornerRadius)
		.modifier(BorderIfNeeded(
			over: document.itsBackgroundColor,
			cornerRadius: geometryGamesCornerRadius))
	}
}


// MARK: -
// MARK: Buttons

struct ResolutionButton: View {
	
	@ObservedObject var document: KaleidoPaintDocument
	@State var showResolutionMenu = false

	var body: some View {
	
		//	The simpler approach of
		//
		//		Menu() {
		//			Picker(selection: $log2UnitCellSize, label: Text("Resolution")) {
		//				...
		//			}
		//		} label: {
		//		}
		//
		//	won't let us show a title and subtitle.
		//	Instead it fishes out the first text it finds
		//	and uses it by itself (thus ignoring the subtitle).
		//	So let's use a popover instead and add the menu functionality
		//	on our own.

		Button() {
			
			showResolutionMenu.toggle()

		} label: {
		
			Image(systemName: "gear")
				.font(.title)
				.padding(geometryGamesControlPadding)
				.background(Color(UIColor.systemBackground))
				.cornerRadius(geometryGamesCornerRadius)
				.modifier(BorderIfNeeded(
							over: document.itsBackgroundColor,
							cornerRadius: geometryGamesCornerRadius))
		}
		.popover(isPresented: $showResolutionMenu) {
		
			VStack(alignment: .leading, spacing: 4.0) {
			
				Text("Resolution")
				.font(.title3)

				ResolutionItem(
					title: "Low Res - Title",
					subtitle: "Low Res - Subtitle",
					value: 9,
					log2UnitCellSize: $document.itsLog2UnitCellResolution,
					showResolutionMenu: $showResolutionMenu
				)

				ResolutionItem(
					title: "Medium Res - Title",
					subtitle: "Medium Res - Subtitle",
					value: 10,
					log2UnitCellSize: $document.itsLog2UnitCellResolution,
					showResolutionMenu: $showResolutionMenu
				)

				ResolutionItem(
					title: "High Res - Title",
					subtitle: "High Res - Subtitle",
					value: 11,
					log2UnitCellSize: $document.itsLog2UnitCellResolution,
					showResolutionMenu: $showResolutionMenu
				)
			}
			.padding(geometryGamesPanelPadding)
		}
	}
}

struct ResolutionItem: View {

	let title: LocalizedStringKey
	let subtitle: LocalizedStringKey
	let value: Int
	@Binding var log2UnitCellSize: Int
	@Binding var showResolutionMenu: Bool

	var body: some View {
	
		Button() {
		
			log2UnitCellSize = value
			
			UserDefaults.standard.set(value, forKey: "default log₂(unit cell size)")
			
			showResolutionMenu = false
		
		} label: {
		
			HStack(alignment: .top) {

				if value == log2UnitCellSize {
				
					Text("✓")
					.font(.body)
					
				} else {
				
					Text("✓")
					.font(.body)
					.hidden()
				}

				VStack(alignment: .leading) {

					Text(title)
					.font(.body)

					Text(subtitle)
					.font(.caption)
					.foregroundColor(.gray)
				}
			}
		}
	}
}

struct UndoButton: View {

	@ObservedObject var document: KaleidoPaintDocument
	
	var body: some View {

		Button {

			document.itsUndoManager?.undo()

			document.itsUndoRedoCount += 1

		} label: {
		
			Image(systemName: "arrowshape.turn.up.backward.circle")
				.font(.title)
		}
		.disabled(!(document.itsUndoManager?.canUndo ?? false || gMakeScreenshots))
	}
}

struct RedoButton: View {

	@ObservedObject var document: KaleidoPaintDocument

	var body: some View {
	
		Button {
			
			document.itsUndoManager?.redo()
			
			document.itsUndoRedoCount += 1
			
		} label: {
		
			Image(systemName: "arrowshape.turn.up.forward.circle")
				.font(.title)
		}
		.disabled(!(document.itsUndoManager?.canRedo ?? false || gMakeScreenshots))
	}
}


struct LockButton: View {
	
	@ObservedObject var document: KaleidoPaintDocument
	let fileURL: URL?
	@Binding var shouldTestForTooThinLinesWhenUnlocked: Bool
	@Binding var autoThickeningMessageShouldBeShown: Bool

	var body: some View {

		let theFileName = String(fileURL?.deletingPathExtension().lastPathComponent ?? "")

		return Menu() {

			if document.itsContentIsLocked {

				Button() {

					document.itsContentIsLocked = false
					
					//	If we haven't already tested for too-thin lines, test now.
					if shouldTestForTooThinLinesWhenUnlocked {
					
						shouldTestForTooThinLinesWhenUnlocked = false
						
						autoThickeningMessageShouldBeShown
							= autoThickenBrushstrokesInLegacyDrawing(document: document)
					}

				} label: {
					Label(
						String(
							format: Bundle.main.localizedString(
										forKey: "UnlockDrawing",
										value: nil,
										table: nil),
							theFileName),
						systemImage: "lock.open.fill")
				}

			} else {

				Button() {
				
					//	Note:  .onChange(of: document.itsContentIsLocked)
					//	will clear the Undo and Redo stacks,
					//	so that 3-finger swipes (which are built into iOS
					//	and can't be disabled as far as I can tell)
					//	will have no effect.
				
					document.itsContentIsLocked = true
					
				} label: {
					Label(
						String(
							format: Bundle.main.localizedString(
										forKey: "LockDrawing",
										value: nil,
										table: nil),
							theFileName),
						systemImage: "lock.fill")
				}
			}

		} label: {

			Image(systemName: (document.itsContentIsLocked ? "lock.fill" : "lock.open.fill"))
				.font(.title)
				.padding(geometryGamesControlPadding)
				.background(Color(UIColor.systemBackground))
				.cornerRadius(geometryGamesCornerRadius)
				.modifier(BorderIfNeeded(
					over: document.itsBackgroundColor,
					cornerRadius: geometryGamesCornerRadius))
		}
	}
}

struct ShadowIfNeeded: ViewModifier {

	@Environment(\.colorScheme) var colorScheme
	
	@Binding var buttonColor: LinearDisplayP3Color
	
	func body(content: Content) -> some View {

		let thePerceivedIntensity
			= 0.299 * buttonColor.r
			+ 0.587 * buttonColor.g
			+ 0.114 * buttonColor.b

		let thePossibleShadowColor: Color?
		
		switch colorScheme {
		
		case .light:
			thePossibleShadowColor = (thePerceivedIntensity > 0.8750 ? Color.gray : nil)
			
		case .dark:
			thePossibleShadowColor = (thePerceivedIntensity < 0.0625 ? Color.white : nil)
			
		@unknown default:
			thePossibleShadowColor = nil
		}

		return Group() {

			if let theShadowColor = thePossibleShadowColor {

				content
				.shadow(color: theShadowColor, radius: 4.0)

			} else {

				content
			}
		}
	}
}

struct BrushColorButton: View {

	@Binding var visiblePanel: PanelType
	@Binding var brushColor: LinearDisplayP3Color

	var body: some View {

		Button {
			visiblePanel = (visiblePanel == .brushColor ? .none : .brushColor)
		} label: {
			Image(systemName: "paintbrush.pointed")
				.font(.title)
				.foregroundColor(brushColor.asSwiftUIColor())
				.modifier(ShadowIfNeeded(buttonColor: $brushColor))
				.padding(geometryGamesControlPadding)
				.background(visiblePanel == .brushColor ?
							selectedButtonBackgroundColor : nil)
				.cornerRadius(selectedButtonCornerRadius)
		}
	}
}

struct FillColorButton: View {

	@Binding var visiblePanel: PanelType
	@Binding var fillColor: LinearDisplayP3Color

	var body: some View {
	
		Button {
			visiblePanel = (visiblePanel == .fillColor ? .none : .fillColor)
		} label: {
			Image("paint bucket")
				.font(.title)
				.foregroundColor(fillColor.asSwiftUIColor())
				.modifier(ShadowIfNeeded(buttonColor: $fillColor))
				.padding(geometryGamesControlPadding)
				.background(visiblePanel == .fillColor ?
							selectedButtonBackgroundColor : nil)
				.cornerRadius(selectedButtonCornerRadius)
		}
	}
}

struct BackgroundColorButton: View {

	@Binding var visiblePanel: PanelType
	@Binding var backgroundColor: LinearDisplayP3Color
	
	var body: some View {
	
		Button {
			visiblePanel = (visiblePanel == .backgroundColor ? .none : .backgroundColor)
		} label: {
			ZStack {
				Image(systemName: "square.fill")
					.imageScale(.large)
					.foregroundColor(backgroundColor.asSwiftUIColor())
				Image(systemName: "square")
					.imageScale(.large)
					.foregroundColor(.gray)
			}
			.font(.title.weight(.ultraLight))
			.padding(geometryGamesControlPadding)
			.background(visiblePanel == .backgroundColor ?
						selectedButtonBackgroundColor : nil)
			.cornerRadius(selectedButtonCornerRadius)
		}
	}
}

struct BrushThicknessButton: View {

	@Binding var visiblePanel: PanelType

	var body: some View {
	
		Button {
			visiblePanel = (visiblePanel == .brushThickness ? .none : .brushThickness)
		} label: {
			Image(systemName: "lineweight")
				.font(.title)
				.padding(geometryGamesControlPadding)
				.background(visiblePanel == .brushThickness ?
							selectedButtonBackgroundColor : nil)
				.cornerRadius(selectedButtonCornerRadius)
		}
	}
}

struct SymmetryButton: View {

	@Binding var visiblePanel: PanelType
	@Binding var wallpaperGroup: WallpaperGroup

	var body: some View {

		Button {
			visiblePanel = (visiblePanel == .symmetry ? .none : .symmetry)
		} label: {
			Image("wallpaper group toolbar buttons/" + wallpaperGroup.rawValue)
				.font(.title)
				.padding(geometryGamesControlPadding)
				.background(visiblePanel == .symmetry ?
							selectedButtonBackgroundColor : nil)
				.cornerRadius(selectedButtonCornerRadius)
		}
	}
}

struct SmoothingToleranceButton: View {

	@Binding var visiblePanel: PanelType

	var body: some View {
	
		Button {
			visiblePanel = (visiblePanel == .smoothingTolerance ? .none : .smoothingTolerance)
		} label: {
			Image(systemName: "scribble")	//	or "alternatingcurrent"
				.font(.title)
				.padding(geometryGamesControlPadding)
				.background(visiblePanel == .smoothingTolerance ?
							selectedButtonBackgroundColor : nil)
				.cornerRadius(selectedButtonCornerRadius)
		}
	}
}

struct ExportButton: View {

	@Binding var visiblePanel: PanelType

	var body: some View {
	

		Button {
			visiblePanel = (visiblePanel == .export ? .none : .export)
		} label: {
			Image(systemName: "square.and.arrow.up")
				.font(.title)
				.padding(geometryGamesControlPadding)
				.background(visiblePanel == .export ?
							selectedButtonBackgroundColor : nil)
				.cornerRadius(selectedButtonCornerRadius)
		}
	}
}

struct HelpButton: View {

	@Binding var visiblePanel: PanelType
	
	let theHelpTypes: Set<PanelType> = [.helpQuickStart, .helpEditor, .helpTutorials, .helpImageGallery, .helpContact, .translators]

	var body: some View {

		if theHelpTypes.contains(visiblePanel) {
		
			//	One of Help menus is already up,
			//	so provide a Button to dismiss it.
			
			Button() {
				visiblePanel = .none
			} label: {
				Image(systemName: "questionmark.circle")
					.font(.title)
					.padding(geometryGamesControlPadding)
					.background(selectedButtonBackgroundColor)
					.cornerRadius(selectedButtonCornerRadius)
			}
			
		} else {
		
			//	None of the Help menus are up,
			//	so provide a Button to let the user
			//	select one.
			
			Menu() {

				Button() {
					visiblePanel = .helpQuickStart
				} label: {
					Label("Quick Start", systemImage: "questionmark.circle")
				}
			
				Button() {
					visiblePanel = .helpEditor
				} label: {
					Label("Editor-HowTo-short", systemImage: "questionmark.circle")
				}
			
				Button() {
					visiblePanel = .helpTutorials
				} label: {
					Label("Tutorials", systemImage: "questionmark.circle")
				}
			
				Button() {
					visiblePanel = .helpImageGallery
				} label: {
					Label("Image Gallery", systemImage: "questionmark.circle")
				}
			
				Button() {
					visiblePanel = .helpContact
				} label: {
					GeometryGamesHelpMenuContactItemLabel()
				}
				
				Button() {
					visiblePanel = .translators
				} label: {
					GeometryGamesHelpMenuTranslatorsItemLabel()
				}
				
			} label: {

				Image(systemName: "questionmark.circle")
					.font(.title)
					.padding(geometryGamesControlPadding)
					.cornerRadius(selectedButtonCornerRadius)
			}
		}
	}
}

struct TouchModePicker: View {

	@ObservedObject var document: KaleidoPaintDocument
	@Binding var visiblePanel: PanelType
	@Binding var editorHelpShouldBeShown: Bool
	@Binding var touchMode: TouchMode
	
	var body: some View {
	
		Picker("Touch Mode", selection: $touchMode) {
		
			ForEach(TouchMode.allCases) { mode in
				Text(LocalizedStringKey(mode.localizedNameKey()))
			}
		}
		.pickerStyle(.segmented)
		.fixedSize()
		.padding(toolbarPadding)
		.background(Color(UIColor.systemBackground))
		.cornerRadius(geometryGamesCornerRadius)
		.modifier(BorderIfNeeded(
			over: document.itsBackgroundColor,
			cornerRadius: geometryGamesCornerRadius))
		.onChange(of: touchMode) { newTouchMode in

			document.itsControlHandleHit = nil

			document.itsBaseDrawingTextureNeedsFullRedraw = true
			document.itsModelChangeCount += 1
			
			//	When entering edit mode...
			switch newTouchMode {
			
			case .draw:

				//	If the helpEditor panel is up,
				//	replace it with the helpQuickStart panel.
				if visiblePanel == .helpEditor {
					visiblePanel = .helpQuickStart
				}

			case .edit:
			
				switch visiblePanel {
				
					//	If the helpEditor panel is up,
					//	leave it up.
					case .helpEditor:
						break
				
					//	If the helpQuickStart panel is up,
					//	replace it with the helpEditor panel.
					case .helpQuickStart:
						visiblePanel = .helpEditor
					
					//	Auto-show helpEditor once per session,
					//	when first switching to Edit mode.
					//	When switching to Edit mode for a second or later time,
					//	close all other panels.
					default:
						if editorHelpShouldBeShown {
							visiblePanel = .helpEditor
						} else {
							visiblePanel = .none
						}
				}
				
				editorHelpShouldBeShown = false
			}
		}
	}
}


// MARK: -
// MARK: Control panels

struct PossibleControlPanel: View {

	@ObservedObject var document: KaleidoPaintDocument
	@Binding var visiblePanel: PanelType
	@Binding var touchMode: TouchMode
	@Binding var brushColorSelectionMode: SelectionMode
	@Binding var brushThicknessSelectionMode: SelectionMode
	@Binding var fillColorSelectionMode: SelectionMode
	@Binding var backgroundColorSelectionMode: SelectionMode
	@Binding var quickStartHelpShouldBeShown: Bool
	let fileURL: URL?
	let renderer: KaleidoPaintRenderer?
	
	//	The brush color, fill color and brush thickness panels
	//	will ignore the changeAllOfSame… bindings that we pass them,
	//	but we have to pass something, and we can't make it a binding
	//	to a 'Bool?' because when the ColorPanel and ThicknessPanel
	//	appear in the EditorPanel, the checkboxes there will need
	//	to be able to toggle the bound values.  So let's define
	//	a dummy Bool value, which will never get used but at least
	//	lets the code compile correctly.
	@State private var dummyBool = false
	

	var body: some View {

		let theTranslatorThanks = [
			GeometryGamesTranslatorThanks(pre: "شكرا لـ", name: "حمزة خليف", post: nil, lang: "ar"),
			GeometryGamesTranslatorThanks(pre: "mit Dank an", name: "Frank Lutz", post: nil, lang: "de"),
			GeometryGamesTranslatorThanks(pre: "gracias a", name: "Joan Porti", post: nil, lang: "es"),
			GeometryGamesTranslatorThanks(pre: "merci à", name: "Roland Lehoucq", post: nil, lang: "fr"),
			GeometryGamesTranslatorThanks(pre: "grazie a", name: "Carlo Petronio", post: nil, lang: "it"),
			GeometryGamesTranslatorThanks(pre: nil, name: "竹内建氏と\n澤田智宏氏に", post: "感謝します", lang: "ja"),
			GeometryGamesTranslatorThanks(pre: "met dank aan", name: "Sebas Eliëns", post: nil, lang: "nl"),
			GeometryGamesTranslatorThanks(pre: "obrigado a", name: "Nuno Fernandes", post: nil, lang: "pt"),
			GeometryGamesTranslatorThanks(pre: "спасибо", name: "Алексею Пронину", post: nil, lang: "ru"),
			GeometryGamesTranslatorThanks(pre: "感谢", name: "马楚媛与邹燕清", post: nil, lang: "zh-Hans")
		]
		

		//	Note:  Some of the panels get left-aligned, others right-aligned.
		switch visiblePanel {
		
		case .none:
		
			Spacer()
			.frame(maxWidth: .infinity, alignment: .center)
		
		case .brushColor:

			ColorPanel(
				document: document,
				colorTarget: .brush,
				colorSelectionMode: $brushColorSelectionMode,
				changeAllOfSameColor: $dummyBool)
			.cornerRadius(geometryGamesCornerRadius)
			.modifier(BorderIfNeeded(
				over: document.itsBackgroundColor,
				cornerRadius: geometryGamesCornerRadius))
			.fixedSize()
			.frame(maxWidth: .infinity, alignment: .leading)
		
		case .fillColor:

			ColorPanel(
				document: document,
				colorTarget: .fill,
				colorSelectionMode: $fillColorSelectionMode,
				changeAllOfSameColor: $dummyBool)
			.cornerRadius(geometryGamesCornerRadius)
			.modifier(BorderIfNeeded(
				over: document.itsBackgroundColor,
				cornerRadius: geometryGamesCornerRadius))
			.fixedSize()
			.frame(maxWidth: .infinity, alignment: .leading)
		
		case .brushThickness:
		
			ThicknessPanel(
				document: document,
				thicknessTarget: .brush,
				thicknessSelectionMode: $brushThicknessSelectionMode,
				changeAllOfSameThickness: $dummyBool)
			.frame(maxWidth: 320.0)
			.cornerRadius(geometryGamesCornerRadius)
			.modifier(BorderIfNeeded(
				over: document.itsBackgroundColor,
				cornerRadius: geometryGamesCornerRadius))
			.frame(maxWidth: .infinity, alignment: .leading)
		
		case .smoothingTolerance:
			
			SmoothingTolerancePanel(document: document)
			.frame(maxWidth: .infinity, alignment: .leading)
		
		case .backgroundColor:
		
			ColorPanel(
				document: document,
				colorTarget: .background,
				colorSelectionMode: $backgroundColorSelectionMode,
				changeAllOfSameColor: $dummyBool)
			.cornerRadius(geometryGamesCornerRadius)
			.modifier(BorderIfNeeded(
				over: document.itsBackgroundColor,
				cornerRadius: geometryGamesCornerRadius))
			.fixedSize()
			.frame(maxWidth: .infinity, alignment: .trailing)

		case .symmetry:

			SymmetryPanel(
				document: document)
			.frame(maxWidth: .infinity, alignment: .trailing)

		case .export:
		
			let theDrawingName = fileURL?.deletingPathExtension().lastPathComponent
		
			if let theRenderer = renderer {
				GeometryGamesExportView(
					modelData: document,
					exportRenderer: theRenderer,
					folderName: theDrawingName,
					imageName: theDrawingName,
					specialSaveTitle: titleForSaveTileableRectangle(shapeConstraints:
						document.itsWallpaperGroup.torusGroup.itsUnitCellShapeConstraints)
				)
				.modifier(BorderIfNeeded(
					over: document.itsBackgroundColor,
					cornerRadius: geometryGamesCornerRadius))
				.frame(maxWidth: .infinity, alignment: .trailing)
			}
			
		case .helpQuickStart:
			
			KaleidoPaintHelpQuickStartView()
			.modifier(BorderIfNeeded(
				over: document.itsBackgroundColor,
				cornerRadius: geometryGamesCornerRadius))
			.frame(maxWidth: .infinity, alignment: .trailing)
			.onAppear() {
				//	We've satisfied our possible obligation
				//	to show the helpQuickStart panel.
				quickStartHelpShouldBeShown = false
			}
			
		case .helpEditor:
			
			KaleidoPaintHelpEditorView()
			.modifier(BorderIfNeeded(
				over: document.itsBackgroundColor,
				cornerRadius: geometryGamesCornerRadius))
			.frame(maxWidth: .infinity, alignment: .trailing)
			
		case .helpTutorials:

			KaleidoPaintHelpTutorialsView()
			.modifier(BorderIfNeeded(
				over: document.itsBackgroundColor,
				cornerRadius: geometryGamesCornerRadius))
			.frame(maxWidth: .infinity, alignment: .trailing)
			
		case .helpImageGallery:
			
			KaleidoPaintHelpImageGalleryView()
			.modifier(BorderIfNeeded(
				over: document.itsBackgroundColor,
				cornerRadius: geometryGamesCornerRadius))
			.frame(maxWidth: .infinity, alignment: .trailing)
			
		case .helpContact:
			
			GeometryGamesContactView()
			.modifier(BorderIfNeeded(
				over: document.itsBackgroundColor,
				cornerRadius: geometryGamesCornerRadius))
			.frame(maxWidth: .infinity, alignment: .trailing)
			
		case .translators:

			GeometryGamesTranslatorsView(theTranslatorThanks)
			.modifier(BorderIfNeeded(
				over: document.itsBackgroundColor,
				cornerRadius: geometryGamesCornerRadius))
			.frame(maxWidth: .infinity, alignment: .trailing)

		}
	}
}

func titleForSaveTileableRectangle(
	shapeConstraints: ShapeConstraints
) -> LocalizedStringKey {

	//	The unit torus is itself rectangular for wallpaper groups
	//	whose ShapeConstraints include .angle_2π_over_4, and
	//	a double cover of the unit torus is rectangular for groups
	//	whose ShapeConstraints include .rhombus.
	//
	//	There are only two groups that fall into neither of those categories,
	//	namely ○ and 2222.  We can export a rectangle even for them,
	//	but it will be tileable only with an offset.
	
	return
		( shapeConstraints.contains(.angle_2π_over_4)
	   || shapeConstraints.contains(.rhombus) )
	?
		"Save Tileable Rectangle"
	:
		"Save Rectangle"
}


enum ColorTarget {
	case brush
	case fill
	case background
	case selectedObject
}

struct ColorPanel: View {

	@ObservedObject var document: KaleidoPaintDocument
	let colorTarget: ColorTarget
	@Binding var colorSelectionMode: SelectionMode
	@Binding var changeAllOfSameColor: Bool
	
	//	WARNING!  Kludge alert:
	//	Dual "sources of truth" for the selected color!
	//
	//	On the one hand, we want to work with LinearDisplayP3Colors
	//	as much as possible, so that in-use colors may get reused
	//	exactly, without introducing the numerical errors that
	//	round-trip conversion to a SwiftUI Color and back would cause.
	@State private var colorPickerColorAsLinearDisplayP3Color: LinearDisplayP3Color
		= LinearDisplayP3Color(r: 0.0, g: 0.0, b: 0.0, a: 0.0)
	//			Note:  .onAppear() will overwrite this color with the correct current color.
	//			It's essential that we set a color here (namely transparent black) that
	//			never appears in any drawing, to ensure that when .onAppear sets
	//			the correct color, the change cascades down through the .onChange(of:) calls
	//			and clears the didJustNowAppear flag.  Otherwise we'd ignore
	//			the first real change the user makes.
	//
	//	On the other hand, the SwiftUI ColorPicker insists
	//	on a binding to a SwiftUI Color.
	@State private var colorPickerColorAsSwiftUIColor: Color
		= Color(.displayP3, red: 0.0, green: 0.0, blue: 0.0, opacity: 0.0) //	overridden in .onAppear()
	//
	//	Conceptually, the colorPickerColorAsLinearDisplayP3Color
	//	is the primary source of truth, and the colorPickerColorAsSwiftUIColor
	//	is a necessary hack.  It makes me cringe to say it,
	//	but when the user explicitly changes the linear P3 color
	//	(for example using the ColorPalette), we copy the change
	//	to the SwiftUI color.  And when the user explicitly changes
	//	the SwiftUI color (using SwiftUI ColorPicker) we copy the change
	//	to the linear P3 color.  The two .onChange(of:) blocks (see below)
	//	do the copying, taking care to avoid an infinite recursion.

	@State private var didJustNowAppear: Bool = true
	

	var body: some View {

		let theOldColor: LinearDisplayP3Color
		let theColorPickerLabel: LocalizedStringKey
		switch colorTarget {
		
		case .brush:
		
			theOldColor = document.itsBrushColor
			theColorPickerLabel = "Brush color"
			
		case .fill:
		
			theOldColor = document.itsFillColor
			theColorPickerLabel = "Fill color"
			
		case .background:
		
			theOldColor = document.itsBackgroundColor
			theColorPickerLabel = "New background color"
			
		case .selectedObject:
		
			if let theControlHandleHit = document.itsControlHandleHit {
			
				let theHitElement = theControlHandleHit.itsElement
				
				theOldColor = theHitElement.itsColor
				
				if theHitElement is Dot {
					theColorPickerLabel = "New dot color"
				} else if theHitElement is Curve {
					theColorPickerLabel = "New line color"
				} else if theHitElement is Fill {
					theColorPickerLabel = "New fill color"
				} else {
					theColorPickerLabel = "???"
				}
				
			} else {	//	should never occur
			
				theOldColor = LinearDisplayP3Color(r: 0.0, g: 0.0, b: 0.0, a: 1.0)
				
				theColorPickerLabel = "???"
			}
		}
		
		let theSwatchCornerRadius = 4.0
		
		return VStack(alignment: .center) {
			
			Picker("Color selection mode", selection: $colorSelectionMode) {
			
				ForEach(SelectionMode.allCases) { mode in
					Image(systemName: mode.symbolSystemName())
				}
			}
			.pickerStyle(.segmented)
			.fixedSize()

			Group() {
			
				//	In terms of the user experience, it would be OK
				//	to keep the ColorPicker available even when we
				//	add the ColorPalette.  But due to the problems
				//	explained in the "Kludge alert" above, this would
				//	lead to problems with the linear P3 color needing
				//	to update the SwiftUI Color with the user selects
				//	an in-use color from ColorPalette, and the
				//	SwiftUI Color needing to update the linear P3 color
				//	when the user selects a new color from the ColorPicker.
				//	To avoid the risk of an infinite recursion,
				//	let's hide the ColorPicker when the ColorPalette
				//	is present.
				//
				switch colorSelectionMode {
							
				case .new:
					ColorPicker(
						theColorPickerLabel,
						selection: $colorPickerColorAsSwiftUIColor,
						supportsOpacity: false)

				case .previous:
					ColorPalette(
						document: document,
						colorPickerColorAsLinearDisplayP3Color:
							$colorPickerColorAsLinearDisplayP3Color,
						colorTarget: colorTarget)
				}
			}
			.frame(maxWidth: .infinity, alignment: .leading)
						
			if colorTarget == .background
			|| colorTarget == .selectedObject {
			
				//	Why an "Apply" button?
				//
				//	As explained on the web page
				//
				//		https://stackoverflow.com/questions/66699101/swiftui-colorpicker-opened-and-closed-events
				//
				//	there's no way to know when the user dismisses
				//	the SwiftUI ColorPicker.  So the only way to put
				//	a color change onto the Undo stack is to insist
				//	that the user explicitly press an Apply button.
				//	This is not an ideal situation, but it's worth
				//	putting up with an Apply button in order to get
				//	the ColorPicker's many good features (custom color palette,
				//	several different selection modes, etc.).
				//
				Button() {
				
					switch colorTarget {
					
					case .brush, .fill:
						//	Changes to the brush and fill colors
						//	are handled above in .onChange(of:)
						//	and don't need an Apply button.
						break
						
					case .background:
					
						document.changeBackgroundColor(
							to: colorPickerColorAsLinearDisplayP3Color)
						
					case .selectedObject:
					
						if let theControlHandleHit = document.itsControlHandleHit {
						
							let theElements = (changeAllOfSameColor ?
								document.itsElements.filter() { element in
									theControlHandleHit.itsElement.itsColor.isAlmostTheSameAs(element.itsColor)
								} :
								[theControlHandleHit.itsElement]
							)
							
							document.changeColor(
								of: theElements,
								to: colorPickerColorAsLinearDisplayP3Color)
						}
					}
				
				} label: {
					Text("Apply")
						.font(.system(size: 20))
				}

				HStack(spacing: 0) {

					//	Swatch of new color
					colorPickerColorAsSwiftUIColor.frame(width: 64, height: 32, alignment: .center)

					//	Swatch of old color
					theOldColor.asSwiftUIColor().frame(width: 64, height: 32, alignment: .center)
				}
				.cornerRadius(theSwatchCornerRadius)
				.modifier(BorderIfNeeded(
					over: LinearDisplayP3Color(fromSwiftUIColor: Color(UIColor.systemBackground)) ??
							LinearDisplayP3Color(r: 1.0, g: 0.0, b: 1.0, a: 1.0),
					cornerRadius: theSwatchCornerRadius))
			}
				
			if colorTarget == .selectedObject {
				Toggle("Change everything of same color", isOn: $changeAllOfSameColor)
					.toggleStyle(.kaleidoPaintCheckbox)
			}
		}
		.font(.body)
		.padding(geometryGamesPanelPadding)
		.background(Color(UIColor.systemBackground))
		.onAppear(){
		
			colorPickerColorAsLinearDisplayP3Color = theOldColor
		}
		.onChange(of: colorPickerColorAsLinearDisplayP3Color) { newColor in
						
			//	Changes to the brush or fill colors are applied immediately,
			//	without putting them onto the Undo stack.
			//
			switch colorTarget {
			
			case .brush:
				document.itsBrushColor = newColor
				
			case .fill:
				document.itsFillColor = newColor
				
			case .background, .selectedObject:
				//	Wait for the user to tap the Apply button.
				break
			}

			//	Kludge alert:
			//
			//	When this ColorPanel appears and the .onAppear() block
			//	sets the linear P3 color, we should copy it to the SwiftUI Color.
			//
			//	Later, if the user selects a linear P3 color from the ColorPalette,
			//	we should likewise copy it to the SwiftUI Color.
			//
			//	Otherwise, assume the ColorPicker changed the SwiftUI Color
			//	and the .onChange(of:) block below called us to update
			//	the linear P3 color accordingly.  Don't send this color back again!
			//
			if didJustNowAppear || colorSelectionMode == .previous {
			
				colorPickerColorAsSwiftUIColor = newColor.asSwiftUIColor()
			}
		}
		.onChange(of: colorPickerColorAsSwiftUIColor) { newColor in
		
			//	Kludge alert:
			//
			//	If the user explicitly choose a new SwiftUI Color
			//	in the ColorPicker, then we should pass it along
			//	to the linear P3 color.
			//
			//	But if the linear P3 color just got set,
			//	either because the ColorPanel just now appeared
			//	or because the user chose an already in-use color
			//	from the ColorPanel, then we don't want
			//	to send that linear P3 color back again!
			
			if didJustNowAppear {

				//	When this ColorPanel first appears, the .onAppear() block (up above)
				//	will set the linear P3 color, and then the .onChange(of:) block
				//	(immediately above) will in turn set the SwiftUI Color accordingly.
				//	Don't send this color back again!
				
				//	Clear the flag.
				didJustNowAppear = false

			} else if colorSelectionMode == .previous {


				//	When the user selects a linear P3 color from the ColorPalette,
				//	the .onChange(of:) block (immediately above) will set
				//	the SwiftUI Color accordingly. Don't send this color back again!

			} else {
			
				//	The user has apparently used the SwiftUI ColorPicker
				//	to change the SwiftUI Color.
				//	Pass it along to the linear P3 color.
				
				colorPickerColorAsLinearDisplayP3Color
					= LinearDisplayP3Color(fromSwiftUIColor: newColor)
					?? LinearDisplayP3Color(r: 0.0, g: 0.0, b: 0.0, a: 1.0)
			}
		}
	}
}

struct ColorPalette: View {

	@ObservedObject var document: KaleidoPaintDocument
	@Binding var colorPickerColorAsLinearDisplayP3Color: LinearDisplayP3Color
	let colorTarget: ColorTarget

	var body: some View {
	
		var theRedundantColors = document.itsElements.map{element in element.itsColor}
		theRedundantColors.append(document.itsBackgroundColor)
		var theIncludedColors = Set<LinearDisplayP3Color>()
		let theUniqueColors = theRedundantColors.filter(){ color in
			theIncludedColors.insert(color).inserted
		}
				
		let theColorChipSize = 32.0
		let theColorChipSpacing: CGFloat? = 0.0
		let theNumColumns = min(6, theUniqueColors.count)
		let theColumns = Array(
			repeating: GridItem(.fixed(theColorChipSize), spacing: theColorChipSpacing),
			count: theNumColumns)
		
		return ScrollView() {
		
			LazyVGrid(columns: theColumns, spacing: theColorChipSpacing) {
			
				//	If we let ForEach() act directly on theUniqueColors,
				//	then we'd need to make LinearDisplayP3Color Identifiable
				//	by adding a "var id: UUID".  That seems like unnecessary
				//	extra weight for what's otherwise a very lightweight struct,
				//	especially since this is the only place we'd need to use it.
				//	So let's let ForEach() act on the array indices instead.
				ForEach(0 ..< theUniqueColors.count, id: \.self) { i in
				
					let theColor = theUniqueColors[i]
					
					//	The web page
					//
					//		https://www.hackingwithswift.com/quick-start/swiftui/how-to-fill-and-stroke-shapes-at-the-same-time
					//
					//	notes that SwiftUI doesn't allow
					//
					//		Rectangle()
					//		.stroke(…)
					//		.fill(…)
					//
					//	and provides this workaround.
					//
					Rectangle()
					.stroke(.black, lineWidth: 2.0)	//	stroke gets centered on rectangle's perimeter
					.background(Rectangle().fill(theColor.asSwiftUIColor()))
					.frame(width: theColorChipSize, height: theColorChipSize)
					.onTapGesture {
						colorPickerColorAsLinearDisplayP3Color = theColor
					}
				}
			}
			.padding(1.0)	//	needed to avoid clipping "strokes" around the perimeter
			.cornerRadius(4.0)
		}
		.frame(maxHeight:
			colorTarget == .selectedObject ?
				2.375 * theColorChipSize :	//	Otherwise in the Editor popover on iPhone,
											//	the ScrollView expands to fill all available space,
											//	even when very little of it is needed.
				.infinity)
	}
}


enum ThicknessTarget {
	case brush
	case selectedObject
}

struct ThicknessPanel: View {

	@ObservedObject var document: KaleidoPaintDocument
	let thicknessTarget: ThicknessTarget
	@Binding var thicknessSelectionMode: SelectionMode
	@Binding var changeAllOfSameThickness: Bool

	//	WARNING!  Kludge alert: Dual "sources of truth" for the targeted thickness!
	//
	//	On the one hand, we want to work with the underlying thickness values
	//	as much as possible, so that in-use thicknesses may get reused exactly,
	//	without introducing the numerical errors that round-trip conversion
	//	to a log₂(thickness factor) and back would cause.
	//
	@State private var plainThickness = gDefaultThickness	//	overridden in .onAppear()
	//
	//	On the other hand, we want the Slider to display the thickness
	//	on a logarithmic scale, so we need to pass the Slider a binding
	//	to an explicit variable for log₂(thickness factor).
	//
	@State private var log2ThicknessFactor = 0.0			//	overridden in .onAppear()
	//
	//	Changes to the plainThickness get converted
	//	and passed along to the log2ThicknessFactor,
	//	except when the user is actively sliding
	//	the Slider, when the reverse is true.
	//
	@State private var userIsSlidingSlider: Bool = false
	

	@State private var startingThickness: Double?	//	Starting thickness for a thickness change
	@State private var endingThickness: Double?		//	 Ending  thickness for a thickness change

	@State private var affectedBrushstrokes: [Brushstroke] = []

	let theDefaultSampleLineThickness = 16.0	//	in points


	var body: some View {

		let theColor: Color
		let theTargetIsDot: Bool
		let theLabel: LocalizedStringKey
		switch thicknessTarget {
		
		case .brush:
		
			theColor = document.itsBrushColor.asSwiftUIColor()
			theLabel = "Brush thickness"
			theTargetIsDot = true
			
		case .selectedObject:
		
			if let theControlHandleHit = document.itsControlHandleHit {
			
				let theSelectedElement = theControlHandleHit.itsElement
				theColor = theSelectedElement.itsColor.asSwiftUIColor()
				theTargetIsDot = theSelectedElement is Dot
				theLabel = (theTargetIsDot ? "Dot size" : "Line thickness")
				
			} else {	//	should never occur
			
				theColor = Color(.displayP3, red: 0.0, green: 0.0, blue: 0.0, opacity: 1.0)
				theTargetIsDot = false
				theLabel = "???"
			}
		}

		let theSampleLineThickness = exp2(log2ThicknessFactor) * theDefaultSampleLineThickness
		let theSampleLineCornerRadius = 0.5 * theSampleLineThickness
	
		return VStack(alignment: .center) {
		
			Picker("Thickness selection mode", selection: $thicknessSelectionMode) {
			
				ForEach(SelectionMode.allCases) { mode in
					Image(systemName: mode.symbolSystemName())
				}
			}
			.pickerStyle(.segmented)
			.fixedSize()

			HStack(alignment: .center) {

				Text(theLabel)
				.fixedSize()
				
				Spacer()
				
				//	The RoundedRectangle() itself has a height
				//	that's proportional to the line thickness.
				//	But it's enclosed in an extra frame of fixed height,
				//	so that the panel's total height stays constant
				//	even as the user varies the line thickness
				//	(well, except for the thickest thicknesses).
				//
				RoundedRectangle(cornerSize: CGSize(
									width: theSampleLineCornerRadius,
									height: theSampleLineCornerRadius))
				.frame(height: theSampleLineThickness)
				.frame(maxWidth: (theTargetIsDot ?
									theSampleLineThickness :
									128.0))
				.foregroundColor(theColor)
				.frame(
					maxWidth: .infinity,
					alignment: (theTargetIsDot ? .center : .trailing))
			}

			switch thicknessSelectionMode {
			
			case .new:

				ThicknessSlider(
					plainThickness: $plainThickness,
					log2ThicknessFactor: $log2ThicknessFactor,
					userIsSlidingSlider: $userIsSlidingSlider,
					startingThickness: $startingThickness,
					endingThickness: $endingThickness)
				
			case .previous:
				
				ThicknessPalette(
					document: document,
					thicknessTarget: thicknessTarget,
					startingThickness: $startingThickness,
					endingThickness: $endingThickness)
			}

			if thicknessTarget == .selectedObject {
				Toggle("Change everything of same thickness", isOn: $changeAllOfSameThickness)
					.toggleStyle(.kaleidoPaintCheckbox)
			}
		}
		.font(.body)
		.padding(geometryGamesPanelPadding)
		.background(Color(UIColor.systemBackground))
		.onAppear{
			
			switch thicknessTarget {
			
			case .brush:
				plainThickness = document.itsBrushThickness
				
			case .selectedObject:
				if let theSelectedBrushstroke
					= document.itsControlHandleHit?.itsElement as? Brushstroke {
					
					plainThickness = theSelectedBrushstroke.itsThickness
					
				} else {	//	should never occur
				
					plainThickness = gDefaultThickness
				}
			}
			
			//	.onChange(of: plainThickness) will now
			//	set log2ThicknessFactor as well.
		}
		.onChange(of: plainThickness) { newPlainThickness in
		
			//	Copy the converted plainThickness to log2ThicknessFactor
			//	in response to .onAppear() and then whenever
			//	the user chooses a color from the ColorPalette.
			//	(When the user is actively sliding the Slider we'll copy
			//	the converted log2ThicknessFactor to the plainThickness.)
			//
			if !userIsSlidingSlider {
			
				log2ThicknessFactor = log2(newPlainThickness / gDefaultThickness)
			}

			//	Whether the user is using the slider or not, apply
			//	the newPlainThickness to any affectedBrushstrokes.
			//
			if thicknessTarget == .selectedObject {

				for theAffectedBrushstroke in affectedBrushstrokes {
				
					//	Set theCurrentThickness without putting it
					//	on the Undo/Redo stack and without recomputing
					//	flood fills.  This will give the user a rough idea
					//	what the new thickness looks like in the drawing.
					
					theAffectedBrushstroke.itsThickness = newPlainThickness
				}
				
				if !affectedBrushstrokes.isEmpty {
				
					document.itsBaseDrawingTextureNeedsFullRedraw = true
					document.itsModelChangeCount += 1
				}
			}
		}
		.onChange(of: log2ThicknessFactor) { newLog2ThicknessFactor in

			//	Copy the converted log2ThicknessFactor to the plainThickness
			//	only while the user is actively sliding the Slider.
			//	(In response to .onAppear(), and then later whenever
			//	the user chooses a color from the ColorPalette we'll copy
			//	the converted plainThickness to the log2ThicknessFactor.)
			//
			if userIsSlidingSlider {
			
				plainThickness = exp2(newLog2ThicknessFactor) * gDefaultThickness
			}
		}
		.onChange(of: startingThickness) { optionalStartingThickness in
		
			if optionalStartingThickness != nil {
			
				switch thicknessTarget {
				
				case .brush:
				
					affectedBrushstrokes = []	//	empty array for brush thickness
				
				case .selectedObject:
				
					if let theSelectedBrushstroke	//	should never fail
						= document.itsControlHandleHit?.itsElement as? Brushstroke {
					
						if changeAllOfSameThickness {
						
							affectedBrushstrokes =
								document.itsElements
								.compactMap() { element in
									element as? Brushstroke
								}
								.filter() { brushstroke in
									brushstroke.itsThickness == theSelectedBrushstroke.itsThickness
								}
						
						} else {
						
							affectedBrushstrokes = [theSelectedBrushstroke]
						}
					
					} else {	//	should never occur
					
						affectedBrushstrokes = []
					}
				}
			}
		}
		.onChange(of: endingThickness) { optionalEndingThickness in
		
			if let theEndingThickness = optionalEndingThickness {
			
				switch thicknessTarget {
				
				case .brush:
				
					//	Set the brush thickness directly,
					//	without putting it on the Undo stack.
					document.itsBrushThickness = theEndingThickness
					
				case .selectedObject:

					if let theStartingThickness = startingThickness {
					
						//	Put the change from the theStartingThickness
						//	to theEndingThickness on the Undo stack.
						//
						//	changeThickness() will change
						//	the affectedBrushstrokes' thicknesses,
						//	and also call clearCachedRLERuns()
						//	to re-create any affected fills.
						
						document.changeThickness(
							of: affectedBrushstrokes,	//	listed in the same order they get drawn
							from: theStartingThickness,
							to: theEndingThickness)
					}
				}
				
				//	Clean up.
				startingThickness = nil
				endingThickness = nil
				affectedBrushstrokes = []	//	OK to clear before setting plainThickness,
											//	because document.changeThickness() will have
											//	taken care of the affectedBrushstrokes.

				//	If the user slid the ThicknessSlider, it will have kept
				//	the log2ThicknessFactor and thence also the plainThickness
				//	updated at all times.
				//
				//	But if the user selected theEndingThickness
				//	in the ThicknessSlider, we'll have to update
				//	the plainThickness and thence also the log2ThicknessFactor.
				//
				if thicknessSelectionMode == .previous {
				
					//	We've already cleared the array of affectedBrushstrokes,
					//	setting plainThickness won't have any side effects
					//	other than to (sometimes redundantly) set log2ThicknessFactor.
					plainThickness = theEndingThickness
				}
			}
		}
	}
}

struct ThicknessSlider: View {

	@Binding var plainThickness: Double
	@Binding var log2ThicknessFactor: Double
	@Binding var userIsSlidingSlider: Bool
	@Binding var startingThickness: Double?	//	Starting thickness for a thickness change
	@Binding var endingThickness: Double?	//	 Ending  thickness for a thickness change

	var body: some View {

		Slider(
			value: $log2ThicknessFactor,
			in: minLog2ThicknessFactor ... maxLog2ThicknessFactor,
			onEditingChanged: {isEditing in
			
				userIsSlidingSlider = isEditing
			
				if isEditing {

					//	The user began sliding the slider.
					//	.onChange(of: startingThickness) will compute the affectedBrushstrokes.
					//
					//		Note:  Apple's documentation says only that
					//
					//			"The slider calls onEditingChanged
					//			when editing begins and ends."
					//
					//		without explicitly say whether the "editing begins" call
					//		includes some small initial change or not.  Fortunately,
					//		in practice the value we get here for log2ThicknessFactor
					//		is the value from just before editing began.  That is,
					//		it doesn't include any small initial change.
					//
					
					startingThickness = plainThickness

				} else {

					//	The user finished sliding the slider.
					//	.onChange(of: endingThickness) will handle the change.
					
					endingThickness = plainThickness
				}
			}
		)
		.frame(width: thicknessSliderLength)
	}
}

//	The length of the SwiftUI Slider includes room
//	for the entire "thumb" at each end.
//	In other words, the usable track length
//	is two thumb radii less than the length
//	of the Slider as a whole.
//
let theExpectedThumbRadius = 13.0	//	an empiral observation -- not documented
let theTrackLength = thicknessSliderLength - 2.0 * theExpectedThumbRadius
let theTrackHeight = 4.0
let theInUseMarkerHeight = 20.0
let theInUseMarkerWidth = 4.0
let theCurrentMarkerHeight = 32.0
let theCurrentMarkerWidth = 2.0
let theButtonHeight = theCurrentMarkerHeight
let theTrackColor = Color(.displayP3, red: 0.875, green: 0.875, blue: 0.875, opacity: 0.875)

struct ThicknessPalette: View {

	@ObservedObject var document: KaleidoPaintDocument
	let thicknessTarget: ThicknessTarget

	@Binding var startingThickness: Double?
	@Binding var endingThickness: Double?
	
	let inUseMarkerColor = Color(.displayP3, red: 0.00, green: 0.75, blue: 0.25, opacity: 1.0)
	let currentMarkerColor = Color(.displayP3, red: 0.00, green: 0.50, blue: 0.75, opacity: 1.0)

	var body: some View {
			
		//	Compute the positions where the in-use thickness markers
		//	should site on the track.  The positions are measured
		//	in points, on a scale from 0 to theTrackLength.
		let theBrushstrokes = document.itsElements.compactMap(){ element in
			element as? Brushstroke}
		let theRedundantThickesses = theBrushstrokes.map() {brushstroke in brushstroke.itsThickness}
		var theIncludedThickesses = Set<Double>()
		let theUniqueThicknesses = theRedundantThickesses.filter(){ thickness in
			theIncludedThickesses.insert(thickness).inserted
		}
		let theSortedThicknesses = theUniqueThicknesses.sorted()
		let thePlacements: [(Double, Double)]
				//	returns (thickness, x position in range (0.0, theTrackLength))
			= theSortedThicknesses.map() { thickness in
			(
				thickness,
				(
					(log2(thickness / gDefaultThickness) - minLog2ThicknessFactor)
				  / (      maxLog2ThicknessFactor        - minLog2ThicknessFactor)
				)
				* theTrackLength
			)
		}
		let theRanges: [(Double, Double, Double)] = makeTappableRanges(placements: thePlacements)
		
		let theCurrentThickness: Double = ( thicknessTarget == .brush ?
			
				//	.brush
				document.itsBrushThickness  :
				
				//	.selectedObject
				(document.itsControlHandleHit?.itsElement as? Brushstroke)?.itsThickness ?? gDefaultThickness
		)
			
		let theCurrentThicknessPosition =
			(
				(log2(theCurrentThickness / gDefaultThickness) - minLog2ThicknessFactor)
			  / (maxLog2ThicknessFactor - minLog2ThicknessFactor)
			)
			* theTrackLength
		
		ZStack() {

			Rectangle()
			.frame(width: theTrackLength, height: theTrackHeight)
			.foregroundColor(theTrackColor)
			.cornerRadius(2.0)
			
			ForEach(thePlacements, id: \.1) { thickness, position in

				Rectangle()
				.frame(width: theInUseMarkerWidth, height: theInUseMarkerHeight)
				.foregroundColor(inUseMarkerColor)
				.cornerRadius(2.0)
				.position(	//	places Rectangle's center
					x: theExpectedThumbRadius + position,
					y: 0.5 * theCurrentMarkerHeight)
			}
			
			Rectangle()
			.frame(width: theCurrentMarkerWidth, height: theCurrentMarkerHeight)
			.foregroundColor(.blue)
			.position(	//	places Rectangle's center
				x: theExpectedThumbRadius + theCurrentThicknessPosition,
				y: 0.5 * theCurrentMarkerHeight)

			//	Invisible -- yet tappable -- rectangles provide
			//	nice wide targets for the user to tap.
			//	Each tappable rectangle includes a given thickness
			//	and extends halfway to the neighboring thickness
			//	on either side.
			//
			ForEach(theRanges, id: \.1) { thickness, leading, trailing in

				Rectangle()
				.frame(width: trailing - leading, height: theCurrentMarkerHeight)
				.foregroundColor(.clear)	//	By default, transparent rectangles aren't tappable, but…
				.contentShape(Rectangle())	//	…this makes rectangle tappable in spite of transparency.
				.position(	//	places Rectangle's center
					x: theExpectedThumbRadius + 0.5 * (leading + trailing),
					y: 0.5 * theCurrentMarkerHeight)
				.onTapGesture() {
				
					//	The ThicknessPanel's onChange(of: startingThickness)
					//	will determine which Brushstrokes are affected.
					startingThickness = theCurrentThickness
					
					//	The ThicknessPanel's onChange(of: endingThickness)
					//	will make the change.
					endingThickness = thickness
				}
			}
		}
		.frame(
			width: thicknessSliderLength,
			height: theCurrentMarkerHeight)
	}

	func makeTappableRanges(
		placements: [(Double, Double)]
	) -> [(Double, Double, Double)] {	//	returns (thickness, leading x, trailing x)
	
		//	Let each range include the given thickness
		//	and extend halfway to the neighboring thickness
		//	on either side.
		
		let theIndices = 0..<placements.count
		let theTappableRanges = theIndices.map() {i in
		
			(
				//	thickness
				placements[i].0,
				
				//	leading x
				i > 0 ?
					0.5 * (placements[i-1].1 + placements[i].1) :
					0.0,
					
				//	trailing x
				i < theIndices.upperBound - 1 ?
					0.5 * (placements[i].1 + placements[i+1].1) :
					theTrackLength
			)
		}
		
		return theTappableRanges
	}
}


struct KaleidoPaintCheckboxToggleStyle: ToggleStyle {

	func makeBody(configuration: Configuration) -> some View {

		HStack() {

			configuration.label
			.font(.system(size: 10))
			//	.multilineTextAlignment(.leading)
			
			Spacer()
			
			Button() {
			
				configuration.isOn.toggle()
				
			} label: {
			
					Image(systemName:
						configuration.isOn ? "checkmark.square" : "square")
			}
			.buttonStyle(.borderless)
		}
	}
}

extension ToggleStyle where Self == KaleidoPaintCheckboxToggleStyle {
	static var kaleidoPaintCheckbox: KaleidoPaintCheckboxToggleStyle { .init() }
}


enum WallpaperGroupNotation: String, CaseIterable, Identifiable {
	
	case none
	case orbifold
	case crystallographic
	
	var id: String { self.rawValue }

	var localizedName: String {
	
		Bundle.main.localizedString(
			forKey: self.rawValue,
			value: nil,
			table: nil)
	}
}

struct SymmetryPanel: View {

	@Environment(\.horizontalSizeClass) var horizontalSizeClass
	@Environment(\.verticalSizeClass) var verticalSizeClass

	@ObservedObject var document: KaleidoPaintDocument
	
	@State var notation = WallpaperGroupNotation(rawValue:
		UserDefaults.standard.string(forKey: "wallpaper group notation") ?? "orbifold") ?? .orbifold

	let theButtonHSpacing = 8.0
	let theButtonVSpacing = 8.0

	var body: some View {

		let theDisplayIsLarge = (
			horizontalSizeClass == .regular
		 && verticalSizeClass   == .regular)
		 
		let theButtonOrderOnCompactDisplay: [WallpaperGroup] =
			[.wpg2222, .wpgS2222, .wpg2S22, .wpg333, .wpgS333, .wpg3S3, .wpg442, .wpgS442, .wpg4S2, .wpg632, .wpgS632, .wpg22S, .wpg22X, .wpgSS, .wpgSX, .wpgXX, .wpgO]
		
		return VStack(alignment: .leading) {

			//	On iOS 15 I couldn't find a way to cajole SwiftUI
			//	to reliably display the title "Symmetry Pattern Longer"
			//	(which simulates the longer title we might expect
			//	in some non-English languages) without truncation
			//	when theDisplayIsLarge == false.  Strangely enough,
			//	it would sometimes split the title into two lines
			//	and sometimes truncate it, depending on seemingly
			//	irrelevant factors.  So as a fallback plan,
			//	let's display a shorter title on small displays.
			//
			Text(theDisplayIsLarge ? "Symmetry pattern" : "Symmetry")
			.padding(EdgeInsets(top: 4.0, leading: 0.0, bottom: 8.0, trailing: 0.0))
			.frame(maxWidth: .infinity, alignment: .center)
		
			if theDisplayIsLarge {
			
				VStack(alignment: .center, spacing: theButtonVSpacing) {
				
					HStack(spacing: theButtonHSpacing) {
					
						WallpaperGroupButton(
							wallpaperGroup: .wpgS2222,
							document: document,
							notation: notation)
							
						WallpaperGroupButton(
							wallpaperGroup: .wpgS333,
							document: document,
							notation: notation)
						
						WallpaperGroupButton(
							wallpaperGroup: .wpgS442,
							document: document,
							notation: notation)
						
						WallpaperGroupButton(
							wallpaperGroup: .wpgS632,
							document: document,
							notation: notation)
					
						WallpaperGroupButton(
							wallpaperGroup: .wpgO,
							document: document,
							notation: notation)
						.hidden()
					
						WallpaperGroupButton(
							wallpaperGroup: .wpgO,
							document: document,
							notation: notation)
						.hidden()
					
						WallpaperGroupButton(
							wallpaperGroup: .wpgO,
							document: document,
							notation: notation)
					}
					
					HStack(spacing: theButtonHSpacing) {
					
						WallpaperGroupButton(
							wallpaperGroup: .wpg2222,
							document: document,
							notation: notation)
						
						WallpaperGroupButton(
							wallpaperGroup: .wpg333,
							document: document,
							notation: notation)
						
						WallpaperGroupButton(
							wallpaperGroup: .wpg442,
							document: document,
							notation: notation)
						
						WallpaperGroupButton(
							wallpaperGroup: .wpg632,
							document: document,
							notation: notation)
					
						WallpaperGroupButton(
							wallpaperGroup: .wpgO,
							document: document,
							notation: notation)
						.hidden()
						
						WallpaperGroupButton(
							wallpaperGroup: .wpg22S,
							document: document,
							notation: notation)
						
						WallpaperGroupButton(
							wallpaperGroup: .wpg22X,
							document: document,
							notation: notation)
					}
					
					HStack(spacing: theButtonHSpacing) {
						
						WallpaperGroupButton(
							wallpaperGroup: .wpg2S22,
							document: document,
							notation: notation)
						
						WallpaperGroupButton(
							wallpaperGroup: .wpg3S3,
							document: document,
							notation: notation)
						
						WallpaperGroupButton(
							wallpaperGroup: .wpg4S2,
							document: document,
							notation: notation)
					
						WallpaperGroupButton(
							wallpaperGroup: .wpgO,
							document: document,
							notation: notation)
						.hidden()
						
						WallpaperGroupButton(
							wallpaperGroup: .wpgSS,
							document: document,
							notation: notation)
						
						WallpaperGroupButton(
							wallpaperGroup: .wpgSX,
							document: document,
							notation: notation)
						
						WallpaperGroupButton(
							wallpaperGroup: .wpgXX,
							document: document,
							notation: notation)
					}
				}
			
			} else {	//	size class is compact
			
				ScrollViewReader() { proxy in
				
					ScrollView(.horizontal) {
					
						HStack() {
						
							ForEach(theButtonOrderOnCompactDisplay) { wallpaperGroup in
							
								WallpaperGroupButton(
									wallpaperGroup: wallpaperGroup,
									document: document,
									notation: notation)
								.padding(2.0)	//	workaround for a clipping artifact in ScrollView
								.id(wallpaperGroup.id)
							}
						}
					}
					.onAppear() {
						proxy.scrollTo(document.itsWallpaperGroup.id,
						anchor: .center)
					}
				}
			}

			AdaptiveStack(horizontalAlignment: .leading, spacing: 0.0) {

				HStack(spacing: 0.0) {
				
					NotationButton(notation: $notation)
					
					Text(":  " + notation.localizedName)
						.font(.footnote)
				}
				
				Spacer()
				
				RealignButton(document: document)
				
				Spacer()

				if document.itsWallpaperGroup.torusGroup.itsRigidity != .rigid {
				
					ShapeButton(document: document)
				}
			}
			
			Divider()
			
			VStack() {
				
				Toggle(isOn: $document.itsShowSingularitiesFlag) {
					Text("Grid")
				}
					
				Toggle(isOn: $document.itsColorBySymmetryFlag) {
					Text("Color by symmetry")
				}
			}
			.fixedSize(
				//	On a large display,
				//		pass horizontal:true to constrain the toggle switches
				//		to the minimum width needed to show their full titles.
				//	On a small display,
				//		pass horizontal:false to allow the parent view
				//		to constrain the toggle switches to their sibling view's width.
				horizontal: theDisplayIsLarge,
				vertical: false)
			.frame(maxWidth: .infinity, alignment: .center)
		}
		.frame(maxWidth: (theDisplayIsLarge ? .infinity : 172.0))
		.fixedSize()
		.padding(geometryGamesPanelPadding)
		.background(Color(UIColor.systemBackground))
		.cornerRadius(geometryGamesCornerRadius)
		.modifier(BorderIfNeeded(
			over: document.itsBackgroundColor,
			cornerRadius: geometryGamesCornerRadius))
		.onChange(of: notation) { newNotation in
		
			UserDefaults.standard.set(
				newNotation.rawValue,
				forKey: "wallpaper group notation")
		}
		.onChange(of: document.itsShowSingularitiesFlag) { _ in
		
			//	We don't need to redraw itsUnitCellSingularitiesTexture,
			//	because we keep it up date even when it's not visible,
			//	but we do need to refresh itsUnitCellComposedTexture
			//	to include it.
			
			document.itsCompositeTextureNeedsRefresh = true
			document.itsModelChangeCount += 1
		}
		.onChange(of: document.itsColorBySymmetryFlag) { _ in
		
			document.itsBaseDrawingTextureNeedsFullRedraw = true
			document.itsCurveInProgressTextureNeedsRedraw = true	//	unnecessary but safe
			document.itsModelChangeCount += 1
		}
	}
}

struct WallpaperGroupButton: View {

	let wallpaperGroup: WallpaperGroup
	@ObservedObject var document: KaleidoPaintDocument
	let notation: WallpaperGroupNotation
	
	var body: some View {
	
		let theOptionalLabel: String?
		switch notation {
		
		case .none:
			theOptionalLabel = nil
			
		case .orbifold:
			theOptionalLabel = wallpaperGroup.orbifoldName
			
		case .crystallographic:
			theOptionalLabel = wallpaperGroup.IUCName
		}
	
		let theButtonSize =	64.0	//	size in points
		let theCornerRadius = 16.0
		let theBorderWidth = 1.0
		
		let theDefaultButtonBackgroundColor = Color.white
		let theSelectedButtonBackgroundColor = Color(.displayP3, red: 0.00, green: 0.25, blue: 1.00)
		let theBorderColor = Color(.displayP3, red: 0.50, green: 0.50, blue: 0.75)
	
		return Button() {
		
			if wallpaperGroup != document.itsWallpaperGroup {
			
				let theDefaultUnitCellPlacement = UnitCellPlacement(
					defaultForWallpaperGroup: wallpaperGroup)

				document.changeWallpaperGroup(
					to: wallpaperGroup,
					placement: theDefaultUnitCellPlacement)
			}
		
		} label: {
		
			VStack(spacing: 4.0) {
		
				Image("wallpaper group panel buttons/" + wallpaperGroup.rawValue)
					.resizable()
					.frame(width: theButtonSize, height: theButtonSize)
					.background(wallpaperGroup == document.itsWallpaperGroup ?
									theSelectedButtonBackgroundColor :
									theDefaultButtonBackgroundColor)
					.cornerRadius(theCornerRadius)
					.overlay(
						RoundedRectangle(cornerRadius: theCornerRadius)
							.stroke(theBorderColor, lineWidth: theBorderWidth)
					)
				
				if let theLabel = theOptionalLabel {
				
					Text(theLabel)
						.font(.body)
				}
			}
		}
	}
}

struct NotationButton: View {

	@Binding var notation: WallpaperGroupNotation

	var body: some View {

		Menu() {
			Picker(
				"Notation",	//	label not shown
				selection: $notation
			) {
				ForEach(WallpaperGroupNotation.allCases, id: \.self) { notationName in
					Text(notationName.localizedName)
				}
			}
		} label: {
			Text("Notation")
				.font(.footnote)
		}
	}
}

struct RealignButton: View {

	var document: KaleidoPaintDocument	//	no need for @ObservedObject

	var body: some View {

		Button() {
			
			//	Restore itsUnitCellPlacement's default position.
			//
			//		Note:  When the user uses two fingers
			//		to translate, rotate and/or zoom the drawing,
			//		we make the change directly, without putting it
			//		onto the Undo stack.  But when the user
			//		taps the Realign button, we *do* put the change
			//		onto the Undo stack, because it might alter
			//		the drawing's aspect ratio.
			//
			
			let theOldPlacement = document.itsUnitCellPlacement
			let theNewPlacement = UnitCellPlacement(
									defaultForWallpaperGroup: document.itsWallpaperGroup)
			
			document.changeUnitCellPlacement(
				oldPlacement: theOldPlacement,
				newPlacement: theNewPlacement)

		} label: {
			Text("Realign")
				.font(.footnote)
		}
	}
}

struct ShapeButton: View {

	@ObservedObject var document: KaleidoPaintDocument
	
	@State var itsOldShape: UnitCellShape?

	var body: some View {

		Button() {
			
			document.itsShapePanelIsShown.toggle()

		} label: {
			Text("Aspect ratio")
				.font(.footnote)
		}
		.popover(isPresented: $document.itsShapePanelIsShown) {
		
			//	The popover seems to be modal only with respect
			//	to taps in Views that don't at least potentially affect
			//	the presence of the ShapeButton or ShapePanel.
			//	By contrast, Views that might affect the presence
			//	of the ShapeButton or ShapePanel respond to the tap used to dismiss
			//	the popover and, when that happens, sometimes
			//	document.itsShapePanelIsShown doesn't get reset to false
			//	and thus document.changeUnitCellShape() doesn't get called.
			//	That happens, for example, when the user selects
			//	a reshapable group (for example ××), opens the ShapePanel
			//	and then taps on a Button for a non-reshapable group
			//	(for example *623) or on a toolbar Button.
			//
			//	To ensure that document.changeUnitCellShape() gets called,
			//	we may respond to .onAppear() and .onDisappear()
			//	and ignore .onChange(of: document.itsShapePanelIsShown).

			ShapePanel(document: document)
			.onAppear() {

				//	The panel is just now appearing.
				//	Remember the unit cell's shape.
				
				itsOldShape = document.getUnitCellShape()
			}
			.onDisappear() {

				//	The panel is just now disappearing.
				//	If unit cell's shape has changed,
				//	submit the change to the Undo stack.
				
				let theShape = document.getUnitCellShape()
				
				if theShape != itsOldShape {
			
					if let theOldShape = itsOldShape {
					
						document.changeUnitCellShape(
							oldShape: theOldShape,
							newShape: theShape)
						
						itsOldShape = nil	//	just to be tidy
					}
				}
			}
		}
	}
}


struct SmoothingTolerancePanel: View {

	@ObservedObject var document: KaleidoPaintDocument

	let theMinSmoothingTolerance = 0.001
	let theMaxSmoothingTolerance = 0.016
	
	let theLabelSize = 32.0
	let theLabelLineThickness = 4.0

	var body: some View {
		
		let theBrushColorAsSwiftUIColor = document.itsBrushColor.asSwiftUIColor()
		
		return VStack(alignment: .center) {
		
			Text("Line smoothing")
		
			Slider(
				value: $document.itsSmoothingTolerance,
				in: theMinSmoothingTolerance ... theMaxSmoothingTolerance,
				label: {
					Text("Smoothing Tolerance")	//	SwiftUI doesn't show the label
				},
				minimumValueLabel: {
					SmoothedLine(lineIsWiggly: true)
					.stroke(style: StrokeStyle(
									lineWidth: theLabelLineThickness,
									lineCap: .round))
					.foregroundColor(theBrushColorAsSwiftUIColor)
					.frame(width: theLabelSize, height: theLabelSize)
				},
				maximumValueLabel: {
					SmoothedLine(lineIsWiggly: false)
					.stroke(style: StrokeStyle(
									lineWidth: theLabelLineThickness,
									lineCap: .round))
					.foregroundColor(theBrushColorAsSwiftUIColor)
					.frame(width: theLabelSize, height: theLabelSize)
				}
			)
			.frame(maxWidth: 384.0)
			.padding(geometryGamesControlPadding)
		}
		.font(.body)
		.padding(geometryGamesPanelPadding)
		.background(Color(UIColor.systemBackground))
		.cornerRadius(geometryGamesCornerRadius)
		.modifier(BorderIfNeeded(
			over: document.itsBackgroundColor,
			cornerRadius: geometryGamesCornerRadius))
	}
}


struct SmoothedLine: Shape {

	let lineIsWiggly: Bool

	func path(in rect: CGRect) -> Path {
	
		//	Note:  A SwiftUI Path's origin sits at the upper left.
		
		//	Choose the control points.
		let c: [CGPoint]
		if lineIsWiggly {
			c = [
				CGPoint(x: 0.110, y: 0.865),
				CGPoint(x: 0.100, y: 0.605),
				CGPoint(x: 0.260, y: 0.510),
				CGPoint(x: 0.255, y: 0.245),
				CGPoint(x: 0.510, y: 0.260),
				CGPoint(x: 0.605, y: 0.085),
				CGPoint(x: 0.865, y: 0.110)
			]
		} else {
			c = [
				CGPoint(x: 0.25, y: 0.125)
			]
		}
		let theNumControlPoints = c.count
		
		//	Compute the points on the curve.
		var p: [CGPoint] = []
		p.reserveCapacity(theNumControlPoints + 1)
		p.append(CGPoint(x: 0.0, y: 1.0))
		for i in 0 ..< (theNumControlPoints - 1) {
			p.append(CGPoint(
				x: 0.5 * (c[i].x + c[i+1].x),
				y: 0.5 * (c[i].y + c[i+1].y)))
		}
		p.append(CGPoint(x: 1.0, y: 0.0))
		let theNumPoints = p.count
		precondition(
			theNumPoints == theNumControlPoints + 1,
			"Internal error: Bezier points and control points aren't consistent")
		
		//	Scale the control point and regular points
		//	to fit the given CGRect.
		let cc = c.map() { unscaledControlPoint in
		
			CGPoint(
				x: rect.origin.x  +  unscaledControlPoint.x * rect.size.width,
				y: rect.origin.y  +  unscaledControlPoint.y * rect.size.height)
		}
		let pp = p.map() { unscaledPoint in
		
			CGPoint(
				x: rect.origin.x  +  unscaledPoint.x * rect.size.width,
				y: rect.origin.y  +  unscaledPoint.y * rect.size.height)
		}
		
		//	Create thePath.
		
		var thePath = Path()
		
		thePath.move(to: pp[0])
		
		for i in 0 ..< (theNumPoints - 1) {
		
			thePath.addQuadCurve(to: pp[i+1], control: cc[i])
		}
		
		return thePath
	}
}


// MARK: -
// MARK: Shape panel

struct ShapePanel: View {

	@ObservedObject var document: KaleidoPaintDocument
	
	@State var itsLog2AspectRatio = 0.0
	let itsMinLog2AspectRatio = -2.0	//	aspect ratio 1/4
	let itsMaxLog2AspectRatio = +2.0	//	aspect ratio  4
	
	@State var itsShear = 0.0
	let itsMinShear = -2.0
	let itsMaxShear = +2.0

	var body: some View {
	
		let theRigidity = document.itsWallpaperGroup.torusGroup.itsRigidity
		
		let theSnapTolerance = 0.125
		
		//	The code here modifies itsUnitCellPlacement directly.
		//	When the user dismisses this ShapePanel, the caller
		//	takes responsibility for putting the overall change
		//	onto the Undo stack.
		VStack() {
		
			//	When the user opens ShapePanel and then switches to a rigid group,
			//	The ShapePanel will be shown briefly for the rigid group
			//	before it disappears.  So we shouldn't display any sort
			//	of error message for a rigid group - just display an empty VStack.

			if theRigidity != .rigid {	//	group has some flexibility
			
				Slider(
					value: $itsLog2AspectRatio,
					in: itsMinLog2AspectRatio...itsMaxLog2AspectRatio,
					onEditingChanged: {isEditing in
					
						//	When the user releases the slider...
						if !isEditing {
						
							//	...with a value close to zero...
							if abs(itsLog2AspectRatio) < theSnapTolerance {
							
								//	...snap exactly to zero
								itsLog2AspectRatio = 0.0
							}
						}
					}
				)
				.frame(width: 256.0)
				.onAppear() {

					let theShape = document.getUnitCellShape()
					
					itsLog2AspectRatio = log2(theShape.aspectRatio)
				}
				.onChange(of: itsLog2AspectRatio) { newLog2AspectRatio in
									
					let theCurrentShape = document.getUnitCellShape()
					let theNewAspectRatio = exp2(newLog2AspectRatio)
					let theNewShape = UnitCellShape(
										aspectRatio: theNewAspectRatio,
										shear: theCurrentShape.shear)
					
					document.setUnitCellShape(theNewShape)
										
					//	As the user drags the slider, we reshape
					//	the drawing using the existing unit cell textures,
					//	because that's a quick operation.
					//
					//	We won't redraw the unit cell textures
					//	until the user dismisses this ShapePanel,
					//	because that can be a slow operation
					//	if the drawing contain lots of filled regions.
					//
					document.itsModelChangeCount += 1
				}
				
				Text("Aspect ratio")
				.font(.caption)
			}
			
			if theRigidity == .fullyFlexible {
			
				Slider(
					value: $itsShear,
					in: itsMinShear...itsMaxShear,
					onEditingChanged: {isEditing in
					
						//	In order to support "half drop" patterns
						//	(as requested by Rebecca Elfast),
						//	when the user releases the slider...
						if !isEditing {
						
							//	...with a value close to an integer or half-integer...
							let theNearestSnapPoint = 0.5 * round(2.0 * itsShear)
							let theDifference = abs(itsShear - theNearestSnapPoint)
							if theDifference < theSnapTolerance {
							
								//	...snap exactly to theNearestSnapPoint.
								itsShear = theNearestSnapPoint
							}
						}
					})
				.frame(width: 256.0)
				.onAppear() {

					let theShape = document.getUnitCellShape()
					
					itsShear = theShape.shear
				}
				.onChange(of: itsShear) { newShear in
									
					let theCurrentShape = document.getUnitCellShape()
					let theNewShape = UnitCellShape(
										aspectRatio: theCurrentShape.aspectRatio,
										shear: newShear)
					
					document.setUnitCellShape(theNewShape)
										
					//	As the user drags the slider, we reshape
					//	the drawing using the existing unit cell textures,
					//	because that's a quick operation.
					//
					//	We won't redraw the unit cell textures
					//	until the user dismisses this ShapePanel,
					//	because that can be a slow operation
					//	if the drawing contain lots of filled regions.
					//
					document.itsModelChangeCount += 1
				}
			
				Text("Shear")
				.font(.caption)
			}
		}
		.padding(16.0)
	}
}


// MARK: -
// MARK: Editor panel

struct EditorSection: ViewModifier {
	func body(content: Content) -> some View {
		content
		.frame(maxWidth: .infinity, alignment: .center)
		.background(Color(UIColor.systemBackground))
		.cornerRadius(geometryGamesCornerRadius)
	}
}

struct EditorPanel: View {

	@ObservedObject var document: KaleidoPaintDocument
	@Binding var colorSelectionMode: SelectionMode
	@Binding var changeAllOfSameColor: Bool
	@Binding var thicknessSelectionMode: SelectionMode
	@Binding var changeAllOfSameThickness: Bool
	@Binding var deleteAllOfSameColor: Bool

	@State private var desiredNodeType: SplineNodeType = .smooth	//	overridden in .onAppear

	var body: some View {

		let thePaddingAndSpacing = 8.0

		var theDeleteButtonTitle: LocalizedStringKey = "Delete ???"
		if document.itsControlHandleHit?.itsElement is Dot {
			theDeleteButtonTitle = "Delete dot"
		}
		if document.itsControlHandleHit?.itsElement is Curve {
			theDeleteButtonTitle = "Delete line"
		}
		if document.itsControlHandleHit?.itsElement is Fill {
			theDeleteButtonTitle = "Delete fill"
		}

		return VStack(spacing: thePaddingAndSpacing) {
		
			ColorPanel(
				document: document,
				colorTarget: .selectedObject,
				colorSelectionMode: $colorSelectionMode,
				changeAllOfSameColor: $changeAllOfSameColor)
			.modifier(EditorSection())

			if document.itsControlHandleHit?.itsElement is Brushstroke {
			
				ThicknessPanel(
					document: document,
					thicknessTarget: .selectedObject,
					thicknessSelectionMode: $thicknessSelectionMode,
					changeAllOfSameThickness: $changeAllOfSameThickness)
				.modifier(EditorSection())
			}
			
			if let theSelectedCurve = document.itsControlHandleHit?.itsElement as? Curve,
			   let theSelectedNode = document.itsControlHandleHit?.itsNode {
			
				let theSpline = theSelectedCurve.itsSpline
				
				VStack(spacing: 12.0) {
				
					if theSelectedNode !== theSpline.itsNodes.first
					&& theSelectedNode !== theSpline.itsNodes.last {
					
						HStack() {
							
							Text("Control point style")
							
							Spacer()

							Picker(
								"Control point style",	//	unused
								selection: $desiredNodeType
							) {
								ForEach(SplineNodeType.allCases, id: \.id) {theNodeType in
									Text(theNodeType.localizedNameKey()).tag(theNodeType)
								}
							}
							.pickerStyle(MenuPickerStyle())
							.onAppear() {

								desiredNodeType = theSelectedNode.itsNodeType
							}
							.onChange(of: desiredNodeType) { newNodeType in

								//	If theSelectedNode.itsNodeType's original value differs
								//	from the desiredNodeType's hard-coded default value,
								//	then onAppear() will set the correct value.
								//	But a side effect of that is that changeonChange(of:)
								//	may get called when the View first appears.
								//	So we must take care to put a SplineNodeType change
								//	onto the Undo stack only if theSelectedNode.itsNodeType
								//	itself has changed.
								
								let theOldNodeType = theSelectedNode.itsNodeType
								
								if newNodeType != theOldNodeType {
								
									document.changeSplineNodeType(
											curve: theSelectedCurve,
											node: theSelectedNode,
											from: theOldNodeType,
											to: newNodeType)
								}
							}
						}
					}
					
					if theSpline.itsNodes.count >= 3 {	//	must keep ≥ 2 nodes

						Button() {
						
							document.deleteSplineNode(
								curve: theSelectedCurve,
								node: theSelectedNode)

							document.itsEditorIsShown = false
						
						} label: {
						
							HStack() {
								Text("Delete control point")
								Spacer()
								Image(systemName: "minus.circle")
							}
						}
					}

					Button() {

						let (theNewNodeA, theNewNodeB)
							= theSelectedCurve.itsSpline.splitNode(theSelectedNode)
						
						document.splitSplineNode(
							curve: theSelectedCurve,
							oldNode: theSelectedNode,
							newNodeA: theNewNodeA,
							newNodeB: theNewNodeB)

						document.itsEditorIsShown = false
					
					} label: {
					
						HStack() {
							Text("Split control point")
							Spacer()
							Image(systemName: "plus.circle")
						}
					}
				}
				.padding(geometryGamesPanelPadding)
				.modifier(EditorSection())
			}
			
			VStack(spacing: 8.0) {
			
				Button() {

					if let theSelectedElement = document.itsControlHandleHit?.itsElement {
					
						let theDeletionPredicate = ( deleteAllOfSameColor ?
							{ (element: DrawingElement) in
								theSelectedElement.itsColor.isAlmostTheSameAs(element.itsColor) } :
							{ (element: DrawingElement) in
								element === theSelectedElement }
						)

						let theRemainingElements = document.itsElements.filter() { element in
														!theDeletionPredicate(element) }
						
						document.changeElementArray(
							from: document.itsElements,
							to: theRemainingElements)
						
						document.itsEditorIsShown = false
					}
					
				} label: {
				
					HStack() {
						Text(theDeleteButtonTitle)
						Spacer()
						Image(systemName: "trash")
					}
					.foregroundColor(Color.red)
				}

				Toggle("Delete everything of same color", isOn: $deleteAllOfSameColor)
					.toggleStyle(.kaleidoPaintCheckbox)
			}
			.padding(EdgeInsets(
				top: 8.0,
				leading: geometryGamesPanelPadding,
				bottom: 8.0,
				trailing: geometryGamesPanelPadding))
			.modifier(EditorSection())
		}
		.padding(thePaddingAndSpacing)
		.background(GeometryReader() { geometry in
		
				//	There's apparently no simple way to set
				//	the color of the little arrow that ties
				//	a popover to its anchor point.
				//	The following workaround (which I found
				//	on the internet) underlays a rectangle
				//	that's slightly larger than the popover's
				//	content area.  It's an ugly hack, but it works.
				
				let thePadding = 128.0
				
				Color(white: 0.875)
					.frame(
						width:  geometry.size.width  + thePadding,
						height: geometry.size.height + thePadding,
						alignment: .center)
					.frame(
						width:  geometry.size.width,
						height: geometry.size.height,
						alignment: .center)
			}
		)
	}
}


//struct KaleidoPaintContentView_Previews: PreviewProvider {
//	static var previews: some View {
//		KaleidoPaintContentView(document: .constant(KaleidoPaintDocument()))
//	}
//}
