//	KaleidoPaintView.swift
//
//	© 2025 by Jeff Weeks
//	See TermsOfUse.txt

import UIKit
import SwiftUI	//	for Binding<>


class KaleidoPaintView: GeometryGamesView<KaleidoPaintDocument>, UIGestureRecognizerDelegate {

	//	Note:  This KaleidoPaintView exists solely
	//	to support UIKit-based touch handling, which
	//	is needed solely because SwiftUI's native gesture handling
	//	is (as of May 2021) still incomplete and buggy.
	//	If Apple someday gets SwiftUI's native gesture handling
	//	to the point that it meets our needs (for example,
	//	to handle rotation gestures and 1-finger dragging
	//	in the same app), then we may use SwiftUI's gesture handling
	//	to replace this legacy UIKit touch handling,
	//	and at that point this KaleidoPaintView class
	//	may be eliminated entirely (instead we'd used
	//	a GeometryGamesView directly). It seems other people
	//	also have tried and failed to combine a rotation
	//	with a 2-finger drag (or ever a 1-finger draw):
	//
	//		https://forums.developer.apple.com/forums/thread/133398
	//
	//	Hmmm... this KaleidoPaintView seems to serve other purposes now,
	//	such as enabling isExclusiveTouch and isMultipleTouchEnabled
	//	and providing extraRenderParameters().

	var itsVisiblePanel: Binding<PanelType>

	enum TouchStatus {
		case none				//	no touches in progress
		case oneFingerDown		//	one finger down, but not yet moved
		case oneFingerMoving	//	one finger down and moving
		case twoFingers			//	two fingers down and perhaps moving
								//		(also allows user to temporarily revert
								//		to one-finger dragging while staying
								//		in twoFingers mode)
	}
	
	var itsTouchStatus: TouchStatus = .none
	var itsTouchMode: Binding<TouchMode>
	
	//	Used only when itsTouchStatus == .oneFingerDown
	var itsInitialTouchPoint: DisplayPoint? = nil

	//	Tiny (unintended!) finger motions get reported as touchesMoved(),
	//	which makes it harder for the user to tap or double-tap
	//	because the tiny unintended finger motion gets mistakenly
	//	interpreted as a very short line, thus blocking proper recognition
	//	of the tap or double-tap.
	//
	//	A tiny intended finger motion may also block proper recognition
	//	of a 2-finger gesture intended to pinch or rotate the drawing,
	//	because a tiny motion of the first finger makes KaleidoPaint
	//	interpret the gesture as a line before the second finger has
	//	touched the screen at all.
	//
	//	To avoid those problems, let's ignore the first finger's motion
	//	until the total distance exceeds itsMotionThresholdToBegin.
	//
	//		Note:  Originally I tried a value of 2.0 points here,
	//		but Francine said that improved her double-tap success rate
	//		only to about 50%.  With a value of 4.0 points here
	//		her double-tap success rate went to 100%.
	//		But when Adam was using the app while holding his iPod
	//		in one hand (which makes for less precise motions than
	//		when using an iPad) he was having trouble getting
	//		a long-press recognized as a tap.  Increasing
	//		itsMotionThresholdToBegin to 8.0 pts solved that problem.
	//
	let itsMotionThresholdToBegin = 8.0	//	in points

	//	If the end of a second tap comes within itsDoubleTapTime seconds
	//	of the end of a first tap, consider them to be a double tap.
	let itsDoubleTapTime = 0.25		//	in seconds
	weak var itsDoubleTapTimer: Timer? = nil
	var itsRawTapLocation: DisplayPoint? = nil


	init(
		frame frameRect: CGRect,
		modelData: KaleidoPaintDocument,
		renderer: KaleidoPaintRenderer,
		visiblePanel: Binding<PanelType>,
		touchMode: Binding<TouchMode>
	) {
		itsTouchMode = touchMode
		itsVisiblePanel = visiblePanel
		
		//	Call super.init().
		//	Among other things, super.init() will create the CAMetalLayer.
		super.init(
			frame: frameRect,
			modelData: modelData,
			renderer: renderer)
		
		isExclusiveTouch = true	//	Ignore all buttons while user is drawing in the main view
		isMultipleTouchEnabled = true
	}
	
	required init(coder: NSCoder) {
		fatalError("init(coder:) has not been implemented")
	}


// MARK: -
// MARK: Rendering

	override func extraRenderParameters() -> Any? {

		return itsTouchMode.wrappedValue
	}


// MARK: -
// MARK: Touch handling

	override func touchesBegan(
		_ touches: Set<UITouch>,
		with event: UIEvent?
	) {
		
		guard let theEvent = event else {
			return
		}
		guard let theEventTouches = theEvent.touches(for: self) else {
			return
		}
		let theNumTouches = theEventTouches.count

		switch itsTouchStatus {
		
		case .none, .oneFingerDown:

			if theNumTouches == 1 {
			
				guard let theTouch = touches.first else {	//	not sure what "first" means for a Set
					return
				}

				itsTouchStatus			= .oneFingerDown
				itsInitialTouchPoint	= touchLocation(theTouch)

			} else {

				let theInitialPoints: [DisplayPoint]
					= theEventTouches.map({ theTouch in touchLocation(theTouch) })

				itsModelData.multiTouchBegan(at: theInitialPoints)

				itsTouchStatus = .twoFingers
			}

		case .oneFingerMoving:

			//	touchesMoved will ignore multi-touch motions
			//	while in .oneFingerMoving state.
			break

		case .twoFingers:

			//	Additional touches are OK.
			//	We'll just temporarily ignore motions
			//	while the number of touches is three or more.
			break
			
		}
	}
	
	override func touchesMoved(
		_ touches: Set<UITouch>,
		with event: UIEvent?
	) {
		
		guard let theEvent = event else {
			return
		}
		guard let theEventTouches = theEvent.touches(for: self) else {
			return
		}
		let theNumTouches = theEventTouches.count

		switch itsTouchStatus {
		
		case .none:
			break
			
		case .oneFingerDown:

			guard let theTouch = touches.first,	//	not sure what "first" means for a Set
				  let theInitialTouchPoint = itsInitialTouchPoint else {
				return
			}

			let theTouchPoint = touchLocation(theTouch)

			let Δx = theTouchPoint.x - theInitialTouchPoint.x
			let Δy = theTouchPoint.y - theInitialTouchPoint.y
			let theDistanceSquared = Δx * Δx  +  Δy * Δy

			//	Ignore small finger motions of less than itsMotionThresholdToBegin,
			//	as explained in the comment accompanying the definition
			//	of itsMotionThresholdToBegin.
			//
			if theDistanceSquared >= itsMotionThresholdToBegin * itsMotionThresholdToBegin {
			
				//	Every call to touchBegan() *must* be followed by a call to touchMoved(),
				//	so every curve will have at least two nodes.
				itsModelData.touchBegan(
								at: theInitialTouchPoint,
								touchMode: itsTouchMode.wrappedValue)
				itsModelData.touchMoved(to: theTouchPoint)

				itsTouchStatus			= .oneFingerMoving
				itsInitialTouchPoint	= nil
			}

		case .oneFingerMoving:

			guard let theTouch = touches.first else {	//	not sure what "first" means for a Set
				return
			}

			let theTouchPoint = touchLocation(theTouch)

			if theNumTouches == 1 {
			
				itsModelData.touchMoved(to: theTouchPoint)
				
			}
			else {
			
				//	If the user starts drawing, but then touches
				//	one or more additional fingers to the touchscreen,
				//	temporarily ignore all touches.  When the user lifts
				//	all fingers but one, then we'll resume drawing.
			}

		case .twoFingers:
		
			let theTouchMotions: [(DisplayPoint, DisplayPoint)]
				= theEventTouches.map({ theTouch in

				//	Technical Note:  previousLocationInView returns
				//	theTouch's last location different from its current location,
				//	even if theTouch hasn't been at that location for a long time.
				//	For present purposes, we want to know theTouch's location
				//	just before this touchesMoved event.
				//
				let theOldLocation = (
					touches.contains(theTouch) ?	//	Did theTouch just now move?
						previousTouchLocation(theTouch) :	//	fetch previous location
						touchLocation(theTouch)				//	touch is sitting still
				)
				let theNewLocation = touchLocation(theTouch)

				return (theOldLocation, theNewLocation)
			})
			
			itsModelData.multiTouchMoved(theTouchMotions)
		}
	}
	
	override func touchesEnded(
		_ touches: Set<UITouch>,
		with event: UIEvent?
	) {
		
		guard let theEvent = event else {
			return
		}
		//	theEvent.touches() includes the touch that's ending.
		guard let theEventTouches = theEvent.touches(for: self) else {
			return
		}
		let theNumTouches = theEventTouches.count
		let theNumTouchesRemaining = theNumTouches - touches.count

		switch itsTouchStatus {
		
		case .none:
			break

		case .oneFingerDown:

			guard let theTouch = touches.first else {	//	not sure what "first" means for a Set
				return
			}
			let theTouchPoint = touchLocation(theTouch)

			itsTouchStatus = .none
			
			switch itsTouchMode.wrappedValue {
			
			case .draw:

				//	In Draw mode we always want to recognize double taps
				//	and pass them through to the KaleidoPaintDocument,
				//	even if a panel is up.  Only single taps can close panels.

				if let theDoubleTapTimer = itsDoubleTapTimer {
				
					//	itsDoubleTapTimer started when we received the first tap,
					//	and now a second tap has arrived within the itsDoubleTapTime.
					//	Dismiss the timer and report the double tap to the KaleidoPaintDocument.

					//	Invalidating theDoubleTapTimer automatically zeroes out
					//	itsDoubleTapTimer, because itsDoubleTapTimer is a weak reference.
					theDoubleTapTimer.invalidate()

					if let theRawTapLocation = itsRawTapLocation {	//	should never fail
						
						itsModelData.doubleTapForDrawing(at: theRawTapLocation)
					}
					
					itsDoubleTapTimer = nil	//	unnecessary for a weak reference,
											//		but makes our intention clear
					itsRawTapLocation = nil

				} else {	//	no double tap timer is active
				
					//	Remember theTouchPoint and wait for this raw tap
					//	to get resolved as either a single tap or a double tap.

					itsRawTapLocation = theTouchPoint
					
					//	Avoid capturing "self" in the Timer callback below.
					let theTouchMode: Binding<TouchMode> = itsTouchMode
					let theVisiblePanel: Binding<PanelType> = itsVisiblePanel
					let theModelData: KaleidoPaintDocument = itsModelData

					itsDoubleTapTimer = Timer.scheduledTimer(
						withTimeInterval: itsDoubleTapTime,
						repeats: false
					) { timer in
					
						//	If seems unlikely that the user would tap once while in .draw mode,
						//	then switch to .edit mode, and then tap a second time,
						//	all within a fraction of a second, but let's check, just to be safe.
						if theTouchMode.wrappedValue == .draw {
						
							//	itsDoubleTapTime has elapsed.
							//	Treat the raw tap as a single tap.

							//	In Draw mode, a single tap (but not a double tap)
							//	will dismiss a panel and not get passed through
							//	to the KaleidoPaintDocument.
							
							if theVisiblePanel.wrappedValue != .none {
							
								theVisiblePanel.wrappedValue = .none
								
							} else {
							
								theModelData.singleTapForDrawing(at: theTouchPoint)
							}
						}
					}
				}
			
			case .edit:
			
				//	In Edit mode we want to recognize a single tap immediately,
				//	so we can display the Edit panel as quickly as possible,
				//	without waiting to see whether a second tap arrives.
				//	As a consequence, both single and double taps will
				//	dismiss panels in Edit mode.

				if itsVisiblePanel.wrappedValue != .none {
				
					itsVisiblePanel.wrappedValue = .none
					
				} else {
				
					itsModelData.singleTapForEditing(at: theTouchPoint)
				}
			}

		case .oneFingerMoving:

			if theNumTouchesRemaining == 0 {
			
				itsTouchStatus = .none

				itsModelData.touchEnded()
			}

		case .twoFingers:

			if theNumTouchesRemaining == 0 {
			
				itsTouchStatus = .none

				itsModelData.multiTouchEnded(touchMode: itsTouchMode.wrappedValue)
			}
			else
			{
				//	Let reshape continue
				//		with one finger being ignored,
				//		with two fingers as a rigid motion,
				//		with three or more fingers being ignored.
			}
		}
	}
	
	override func touchesCancelled(
		_ touches: Set<UITouch>,
		with event: UIEvent?
	) {
		
		//	With iOS 9 split-screen multitasking, when the user
		//	drags his/her finger near the center of the divider bar
		//	(typically but not necessarily to change the screen division)
		//	the touch gets cancelled.

		switch (itsTouchStatus)
		{
			case .none:	//	Should never occur
			
				break

			case .oneFingerDown:
			
				//	Nothing to cancel
				break

			case .oneFingerMoving:
			
				itsModelData.touchCancelled()

			case .twoFingers:
			
				itsModelData.multiTouchCancelled(touchMode: itsTouchMode.wrappedValue)
		}

		//	Tidy up.
		itsTouchStatus = .none
		itsInitialTouchPoint = nil
	}
	
	private func touchLocation(_ touch: UITouch) -> DisplayPoint {

		//	No need to use preciseLocation(in:).
		//	Even with ordinary location(in:), when the user
		//	draws a curve we're going to get way too many raw points,
		//	most of which we'll need to discard before constructing
		//	the spline.
		
		let theCGPoint = touch.location(in: self)
		let theDisplayPoint = convertCGPointToDisplayPoint(theCGPoint)
		
		return theDisplayPoint
	}
	
	private func previousTouchLocation(_ touch: UITouch) -> DisplayPoint {

		//	No need to use precisePreviousLocation(in:).
		//	See comment in touchLocation() immediately above.

		let theCGPoint = touch.previousLocation(in: self)
		let theDisplayPoint = convertCGPointToDisplayPoint(theCGPoint)
		
		return theDisplayPoint
	}
	
	private func convertCGPointToDisplayPoint(_ point: CGPoint) -> DisplayPoint {

		var thePoint = point
		
		//	Measure relative to view's upper-left corner
		//	(which may be non-zero if the view is contained
		//	in a navigation controller).
		thePoint.x -= bounds.origin.x
		thePoint.y -= bounds.origin.y

		//	Flip from iOS's top-down coordinates
		//	to our DisplayPoint's bottom-up coordinates.
		thePoint.y = bounds.size.height - thePoint.y
		
		//	Package up the result.
		let theDisplayPoint = DisplayPoint(
			x: Double(thePoint.x),
			y: Double(thePoint.y),
			viewWidth: Double(bounds.size.width),
			viewHeight: Double(bounds.size.height))
		
		return theDisplayPoint
	}
}
